'''
Convenience functions to construct empirical compounds via khipu;
and retrieve, filter and search empCpds.
'''

import json
import numpy as np
from .search import find_all_matches_centurion_indexed_list
from khipu.epdsConstructor import epdsConstructor
from khipu.utils import adduct_search_patterns, \
                            adduct_search_patterns_neg, \
                                isotope_search_patterns, \
                                    extended_adducts


def get_khipu_epds_from_list_peaks(
                    list_peaks, 
                    mode, 
                    isotope_search_patterns,
                    adduct_patterns,
                    extended_adducts,
                    mz_tolerance_ppm,
                    rt_tolerance):
    '''
    Wrapper function to build empirical compounds from a list of JSON features/peaks.

    Parameters
    ----------
    list_peaks : list of features, [{'mz': 133.097023, 'rtime': 654, 'height': 14388.0, 'id': 555}, ...]
    mode : ionization mode, 'pos' or 'neg'.
    isotope_search_patterns : E.g. [ (1.003355, '13C/12C', (0, 0.8)), (2.00671, '13C/12C*2', (0, 0.8)),]
    adduct_patterns : e.g. [ (21.9820, 'Na/H'), (41.026549, 'Acetonitrile')]
    extended_adducts : adducts used for 2nd round search, e.g. [(117.02655, '-NH3'),
                            (17.02655, 'NH3'), (-18.0106, '-H2O'), ...]
    mz_tolerance_ppm : ppm tolerance in examining m/z patterns.
    rt_tolerance : tolerance of retention time, arbitrary unit but consistent with list_peaks.

    Returns
    -------
    dict_empCpds : {interim_id: empCpd, ...}
    '''
    ECCON = epdsConstructor(list_peaks, mode=mode)
    return ECCON.peaks_to_epdDict(
                    isotope_search_patterns,
                    adduct_patterns,
                    extended_adducts,
                    mz_tolerance_ppm,
                    rt_tolerance,
    ) 

def load_epds_from_json(file):
    '''Read JSON annotation dictionary from asari/khipu.
    Returns list of empCpds.
    '''
    epds = json.load(open(file))
    return list(epds.values())


def filter_epds(list_epds, neutral_formula_mass=True, multiple_ions=True, C13=True):
    '''Filter list of empCpds by neutral_formula_mass or by multiple_ions.
    list_epds : [{'interim_id': 'kp100_128.0951', 'neutral_formula_mass': 128.09508427175538, 'neutral_formula': None, 
        'Database_referred': [], 'identity': [], 'MS1_pseudo_Spectra': [
        {'apex': 653, 'peak_area': 8047955, 'height': 988673, 'left_base': 642, 'right_base': 659, 
        'goodness_fitting': 0.9397139863801345, 'cSelectivity': 0.3497536945812808, 
        'parent_masstrack_id': 880, 'mz': 130.10569763183594, 'snr': 11, 'id_number': 'F1444', 
        'rtime': 125.24219400000001, 'rtime_left_base': 123.24607800000001, 'rtime_right_base': 126.27832200000002, 
        'representative_intensity': 8047955, 'id': 'F1444', 'isotope': '13C/12C', 
        'modification': 'M+H+', 'ion_relation': '13C/12C,M+H+', 'parent_epd_id': 'kp100_128.0951'}, 
        {'apex': 654, 'peak_area': 84934827, 'height': 11980759, 'left_base': 648, 'right_base': 659, 
        'goodness_fitting': 0.993842814809182, 'cSelectivity': 0.43617021276595747, 
        'parent_masstrack_id': 858, 'mz': 129.10237884521484, 'snr': 3, 'id_number': 'F1261', 
        'rtime': 125.41953000000001, 'rtime_left_base': 124.33713, 'rtime_right_base': 126.27832200000002, 
        'representative_intensity': 84934827, 'id': 'F1261', 'isotope': 'M0', 
        'modification': 'M+H+', 'ion_relation': 'M0,M+H+', 'parent_epd_id': 'kp100_128.0951'}], 
        'MS2_Spectra': []}, 
        ...]
    '''
    result = list_epds
    if neutral_formula_mass:
        result = [x for x in result if x['neutral_formula_mass']]
    if multiple_ions:
        result = [x for x in result if len(x['MS1_pseudo_Spectra'])>1]
    if C13:
        result = [x for x in result if check_13C_M1(x)]
    return result

def check_13C_M1(empCpd):
    '''Check presence of pair of M1 13C ion and the M0 12C ion.
    '''
    ions = [x['isotope'] for x in empCpd['MS1_pseudo_Spectra'] if 'isotope' in x]
    if '13C/12C' in ions and 'M0' in ions:
        return True
    else:
        return False

def get_neutrals(list_epds):
    neutrals = []
    for v in list_epds:
        p = {}
        p['id'] = v['interim_id']
        p['mz'] = v['neutral_formula_mass']
        p['rtime'] = np.mean([x['rtime'] for x in v['MS1_pseudo_Spectra']])
        neutrals.append(p)
    return neutrals

def get_match(cpds, mztree, ppm=5):
    '''Find matches of a list of cpds in mztree.
    cpds : [{'id': 'C00025', 'mw': 147.0532, 'name': 'L-Glutamate'}, ...]
    '''
    match = []
    for x in cpds:
        if 'mw' in x and x['mw']:
            _m = find_all_matches_centurion_indexed_list(x['mw'], mztree, ppm)
            if _m:
                match.append( (x, _m) )
    return match



#
# ------------------------------------------------------------------------------------------
# functions to summarize khipus
#


def get_feature_of_max_intensity(featureList):
    '''
    To get feature of max intensity, and avoid errors by sorting.
    e.g. sorted(M0, reverse=True)[0][1] leas to `TypeError: '<' not supported between instances of 'dict' and 'dict'`
    Use np.argmax here, which is okay with ties.
    '''
    ints = [f['representative_intensity'] for f in featureList]
    idx = np.argmax(ints)
    return featureList[idx]

def get_M0(MS1_pseudo_Spectra):
    '''returns M0 feature with highest representative_intensity.
    Without verifying which ion form.'''
    M0 = [f for f in MS1_pseudo_Spectra if f['isotope']=='M0']
    if M0:
        return get_feature_of_max_intensity(M0)
    else:
        return []
    
def get_M1(MS1_pseudo_Spectra):
    '''returns M+1 feature with highest representative_intensity.
    Without verifying which ion form.'''
    M = [f for f in 
          MS1_pseudo_Spectra if f['isotope']=='13C/12C']
    if M:
        return get_feature_of_max_intensity(M)
    else:
        return []
     
def get_highest_13C(MS1_pseudo_Spectra):
    '''returns 13C labeled feature with highest representative_intensity.
    Without verifying which ion form. Because the label goes with sepecific atoms depending on pathway.
    '''
    M = [f for f in 
          MS1_pseudo_Spectra if '13C/12C' in f['isotope']]
    if M:
        return get_feature_of_max_intensity(M)
    else:
        return []

    
def filter_khipus(list_empCpds, natural_ratio_limit=0.5):
    '''
    returns 
    isopair_empCpds = with good natural 13C ratio, based on M1/M0, not checking adduct form.
 
    Usage
    -----
    full_list_empCpds  = json.load(open(json_empcpd))
    isopair_empCpds = filter_khipus(full_list_empCpds)

    '''
    isopair_empCpds = []
    for epd in list_empCpds:
        M0, M1 = get_M0(epd['MS1_pseudo_Spectra']), get_M1(epd['MS1_pseudo_Spectra'])
        if M0 and M1:
            if float(M1['representative_intensity'])/(1 + float(M0['representative_intensity'])) < natural_ratio_limit:
                isopair_empCpds.append( epd['interim_id'] )

    return isopair_empCpds
    

def filter_khipus_by_samples(kdict, unlabelled=[1, 2, 3], labeled=[0, 4, 5], natural_ratio_limit=0.5):
    '''
    Enumerate khipus with good natural ratio and increased isotope labeling ratio.
    
    kdict : khipu_dict as input
    
    returns 
    list of khipus with good natural 13C ratio, list of khipus with increased 13C ratio.

    '''
    khipus_good_natural_ratios, khipus_increased_ratios = [], []
    for k,v in kdict.items():
        # interim_id = v['interim_id']
        M0, M1, Mx = get_M0(v['MS1_pseudo_Spectra']), get_M1(v['MS1_pseudo_Spectra']
                                            ), get_highest_13C(v['MS1_pseudo_Spectra'])
        if M0 and M1:
            try:
                unlabelled_13C = np.array(M1['intensities'])[unlabelled].mean()
                unlabelled_12C = np.array(M0['intensities'])[unlabelled].mean()
                ratio_natural = unlabelled_13C/(unlabelled_12C+0.1)
                if ratio_natural < natural_ratio_limit:
                    khipus_good_natural_ratios.append( (k, ratio_natural) )
                    if Mx:
                        labelled_13C = np.array(Mx['intensities'])[labeled].mean()
                        labelled_12C = np.array(M0['intensities'])[labeled].mean()
                        ratio_labeled = labelled_13C/(labelled_12C+0.1)
                        if ratio_labeled > ratio_natural:
                            khipus_increased_ratios.append( (k, ratio_labeled) )
            except IndexError:
                pass
    print(len(khipus_good_natural_ratios), len(khipus_increased_ratios))
    return khipus_good_natural_ratios, khipus_increased_ratios
    
    
    
def get_isopairs_good_khipus(list_empCpds, natural_ratio_limit=0.5):
    '''
    returns 
    Two lists of khipus, isopair_empCpds_ids (IDs only), good_khipus (full dict), and number of isopair_mtracks.
    isopair_empCpds = with good natural 13C ratio, based on M1/M0, not checking adduct form.
    good_khipus = isopair_empCpds and M0 being a good feature.
    
    Some inline MS/MS expts cause split MS1 peaks. Thus isopair_mtracks are more indicative of the data coverage.
    
    Usage
    -----
    full_list_empCpds  = json.load(open(json_empcpd))
    isopair_empCpds, num_isopair_mtracks, good_khipus = get_isopairs_good_khipus(full_list_empCpds)

    Use filter_khipus if not considering is_good_peak.
    '''
    isopair_empCpds_ids, isopair_mtracks, good_khipus = [], [], []
    for epd in list_empCpds:
        if len(epd['MS1_pseudo_Spectra']) > 1:
            try:
                M0, M1 = get_M0(epd['MS1_pseudo_Spectra']), get_M1(epd['MS1_pseudo_Spectra'])
                feature_M0 = [f for f in epd['MS1_pseudo_Spectra'] if f['isotope']=='M0'][0]
                if M0 and M1:
                    if float(M1['representative_intensity'])/(1 + float(M0['representative_intensity'])) < natural_ratio_limit:
                        isopair_empCpds_ids.append( epd['interim_id'] )
                        if feature_M0['is_good_peak']:
                            good_khipus.append( epd )
                            isopair_mtracks.append( epd["MS1_pseudo_Spectra"][0]['parent_masstrack_id'] )
            except KeyError:
                print(epd['interim_id'])
                
    return isopair_empCpds_ids, len(set(isopair_mtracks)), good_khipus
    
    
def count_singletons(list_empCpds):
    return len([epd for epd in list_empCpds if len(epd['MS1_pseudo_Spectra'])==1])
    
    
def get_isopair_features(full_list_empCpds, isopair_empCpds):
    '''
    Not clean, including more features than desired..
    '''
    isopair_features = []
    for epd in full_list_empCpds:
        if epd['interim_id'] in isopair_empCpds:
            isopair_features += epd['MS1_pseudo_Spectra']

    return isopair_features
    


'''
This module implements the Acquisition object which is a set of data collected from a sample.

A sample in this case could mean a biologically-derived sample or a blank or any other unit that 
is analyzed for analysis.

Each analytical replicate is therefore its own acquisition.
'''

import os
import numpy as np
import pymzml
import matplotlib.pyplot as plt

from intervaltree import IntervalTree
from metDataModel.core import Sample
from .utils import recursive_encoder


class Acquisition(Sample):
    """
    The Acquisition object represents a single LC-MS run.
    """

    def __init__(
        self,
        name,
        source_filepath=None,
        metadata_tags=None,
        raw_filepath=None,
        mzml_filepath=None,
        ionization_mode=None,
        has_ms2=None,
        experiment=None
    ):
        
        super().__init__(
            experiment = '',
            registry = {
                "input_file": source_filepath,
                "name": name,
                "sample_id": name,
                "list_retention_time": None
            },
            mode = ionization_mode,
            sample_type = '',
            input_file = source_filepath,
            name = name,
            id = name,
        )
        self.experiment = experiment

        self.metadata_tags = metadata_tags if metadata_tags is not None else {}
        self.raw_filepath = raw_filepath
        self.mzml_filepath = mzml_filepath

        #lazily_evaluated
        self.__ionization_mode = ionization_mode
        self.__has_ms2 = has_ms2

    @property
    def source_filepath(self):
        return self.input_file
    
    @staticmethod
    def load_acquisition(acquisition_data, experiment):
        """
        This takes a dict of acquisition data and returns the Acquisition object

        Args:
            acquisition_data (dict): acquisition data

        Returns:
            object: an acquisition object
        """
        return Acquisition(
            acquisition_data["name"],
            source_filepath=acquisition_data["registry"]["input_file"],
            metadata_tags=acquisition_data["metadata_tags"],
            raw_filepath=acquisition_data["raw_filepath"],
            mzml_filepath=acquisition_data["mzml_filepath"],
            ionization_mode=acquisition_data["_Acquisition__ionization_mode"],
            has_ms2=acquisition_data["_Acquisition__has_ms2"],
            experiment=experiment
        )

    @staticmethod
    def create_acquisition(name, source_filepath, metadata_dict, experiment=None):
        """
        This is the primary constructor the acquisition object.

        :param name: the name of the acquisition
        :param source_filepath: the original location of the data file
        :param metadata_dict: the metadata associated with the file,
            provided by the .csv file.

        :return: Acquisition object
        """
        return Acquisition(
            name,
            source_filepath,
            metadata_dict,
            raw_filepath=None,
            mzml_filepath=None,
            ionization_mode=None,
            has_ms2=None,
            experiment=experiment
        )

    @property
    def ionization_mode(self):
        """
        This method determines the ionization mode of the acquisition

        :return: ionization mode, "pos" or "neg"
        """
        if self.__ionization_mode is None:
            try:
                for spec in pymzml.run.Reader(self.mzml_filepath):
                    if spec["positive scan"]:
                        self.__ionization_mode = "pos"
                        self.mode = self.__ionization_mode
                        return self.__ionization_mode
                    self.__ionization_mode = "neg"
                    self.mode = self.__ionization_mode
                    return self.__ionization_mode
            except:
                return self.__ionization_mode
        return self.__ionization_mode

    @property
    def json_repr(self):
        """
        This generates the dict representation of the acquisition, this is used when the experiment is saved or loaded.

        :return: a JSON-friendly dictionary for serialization during experiment saving and loading
        """
        return recursive_encoder({k: v for k, v in self.__dict__.items() if k != 'experiment'})

    @property
    def has_ms2(self):
        """
        Scan the mzml to detect if there are MS2 spectra

        :return: has_MS2, True or False
        """
        method_field = "Method"
        if self.__has_ms2 is None:
            ms_method = None
            if method_field in self.metadata_tags:
                ms_method = self.metadata_tags[method_field]
                if ms_method in self.experiment.MS2_methods:
                    self.__has_ms2 = True
                    return self.__has_ms2
                if ms_method in self.experiment.MS1_only_methods:
                    self.__has_ms2 = False
                    return self.__has_ms2
            fp_to_read = None
            if self.mzml_filepath:
                fp_to_read = self.mzml_filepath
            elif self.source_filepath.endswith(".mzML"):
                fp_to_read = self.source_filepath
            if fp_to_read:
                self.__has_ms2 = False
                try:
                    reader = pymzml.run.Reader(fp_to_read)
                    for _, spec in enumerate(reader):
                        if spec.ms_level == 2:
                            self.__has_ms2 = True
                            break
                except:
                    pass
            if ms_method and self.__has_ms2:
                self.experiment.MS2_methods.add(ms_method)
            elif ms_method and not self.__has_ms2:
                self.experiment.MS1_only_methods.add(ms_method)
        return self.__has_ms2

    def TIC(self, mz=None, ppm=5, rt=None, rt_tol=2, title=None):
        """
        This method generates TIC plots for the acquisition. If mz and rt is not provided, 
        this will make the TIC including the entire rt range and all mz values. If mz and rt values
        are provided as co-indexed lists, then only those regions will be used IN ADDITION to 
        the entire rt and mz range. These are saved as figures. 

        Args:
            mz (list, optional): mz values to limit the TIC calculation to. Defaults to None.
            ppm (int, optional): mass tolerance in ppm for the mz values. Defaults to 5.
            rt (list, optional): rt values to limit the TIC caclulation to. Defaults to None.
            rt_tol (int, optional): rt_tolerance in seconds. Defaults to 2.
            title (string, optional): if provided, sets the title on TIC plot. Defaults to None.

        Returns:
            str: path to the TIC plot
        """        
        if mz is None:
            mz = []
        if rt is None:
            rt = []
        title = (
            self.name
            + ",".join([str(x) for x in mz])
            + "_"
            + ",".join([str(x) for x in rt])
            + "_"
            + str(ppm)
            + "_"
            + str(rt_tol)
        )
        fig_path = os.path.join(
            os.path.abspath(self.experiment.experiment_directory),
            "TICs/",
            title + ".png",
        )
        if os.path.exists(fig_path):
            return fig_path
        os.makedirs(os.path.dirname(fig_path), exist_ok=True)
        mz_trees = [IntervalTree()] + [IntervalTree() for _ in mz]
        rt_trees = [IntervalTree()] + [IntervalTree() for _ in mz]
        mz_trees[0].addi(-np.inf, np.inf)
        rt_trees[0].addi(-np.inf, np.inf)
        for i, (x, y) in enumerate(zip(mz, rt)):
            mz_trees[i + 1].addi(x - x / 1e6 * ppm, x + x / 1e6 * ppm)
            rt_trees[i + 1].addi(y - rt_tol, y + rt_tol)
        bins = [[] for _ in mz_trees]
        rtimes = []
        for spec in pymzml.run.Reader(self.mzml_filepath):
            rtime = round(spec.scan_time[0] * 60, 3)
            rtimes.append(rtime)
            for b in bins:
                b.append(0)
            matches = [bool(rt_tree.at(rtime)) for rt_tree in rt_trees]
            match_mask = [i for i, match in enumerate(matches) if match]
            if match_mask:
                for peak in spec.peaks("centroided"):
                    mz = peak[0]
                    for match in match_mask:
                        if mz_trees[match].at(mz):
                            bins[match][-1] += float(peak[1])
        fig = plt.figure()
        for i, b in enumerate(bins):
            ax = fig.add_subplot(len(b), 1, i + 1)
            ax.plot(rtimes, b)
        plt.savefig(fig_path)
        plt.close()
        return fig_path

    def filter(self, user_filter):
        """
        This method filters acquisition based on their metadata keys.

        The filter is organized as follows::

            {
                "key_1": {
                    "includes": ["substr1", "substr2"],
                    "lacks": ["substr3"]
                }
                ...
            }

        In this case, key_1 must be a field in the metadata. It will pass the filter if and only if
        every substring (substr1, substr2) from includes is present in the metadata field's value AND
        every substring in the lacks field (substr3) is not present in the field's data.

        Multiple keys can be specified in the filter. The results from the filter are AND'd for every
        key.

        :param filter: dictionary as described above
        :return: true if acquisition passed filter else false
        """
        passed_filter = True
        if user_filter:
            for key, rules in user_filter.items():
                values_to_filter = self.metadata_tags[key].strip()
                if "includes" in rules:
                    for must_include in rules["includes"]:
                        passed_filter = (
                            passed_filter and must_include in values_to_filter
                        )
                if "lacks" in rules:
                    for not_include in rules["lacks"]:
                        passed_filter = (
                            passed_filter and not_include not in values_to_filter
                        )
                if "equals" in rules:
                    for must_include in rules["equals"]:
                        passed_filter = (
                            passed_filter and must_include == values_to_filter
                        )
        return passed_filter
    


"""
This is the main module in the pcpfm. All functions that are intended to be called by 
an end user are located here, although API access to the underlying modules is possible.

Each function in the Main object maps to a single command on the command line. 
"""

import os
import json
import multiprocessing as mp
import argparse
import csv
import zipfile
import gdown
import requests
import tarfile
from . import Experiment
from . import EmpCpds
from . import default_parameters
from . import Report

class Main():
    """
    This is simply a wrapper around all the CLI functions. By putting them in 
    this object, we can do clever things with getattr()
    """

    @staticmethod
    def process_params():
        """
        This process parses the command line arguments and returns the 
        parameters in a dictionary. Default parameters are specified in 
        the example_parameters.py file and some are read dynamically from
        .json files as specified in that file. Note that any parameters 
        given as .json files will be assumed to be a file path to a json
        file and read as such. This allows complex datastructures to be 
        specified for some parameters. 

        :return: parameters dictionary
        """
        all_subcommands = "\n".join([m for m in dir(Main) if not m.startswith('__')])
        with open(os.path.join(os.path.dirname(__file__), '__init__.py'), encoding='utf-8') as f:
            for x in f.readlines():
                if '__version__' in x:
                    version = "v" + x.split("=")[-1].strip()[1:-1]
        params = default_parameters.PARAMETERS
        parser = argparse.ArgumentParser(description='pcpfm, LC-MS end-to-end processing', prog='pcpfm')
        parser.add_argument('--version', action='version', version="pcpfm " + version)
        parser.add_argument('subcommand', metavar='subcommand', 
                            help='one of the subcommands: ' + all_subcommands)
        parser.add_argument('-p', '--parameters')
        parser.add_argument('-m', '--mode', default=None)
        parser.add_argument('--ppm', default=5, type=int)
        parser.add_argument('-s', '--sequence')
        parser.add_argument('-c', '--cores', type=int)
        parser.add_argument('-MS2_dir')
        parser.add_argument('-f', '--filter')
        parser.add_argument('-j', '--project')
        parser.add_argument('-o', '--output')
        parser.add_argument('-i', '--input')
        parser.add_argument('--name_field', default='File Name')
        parser.add_argument('--path_field', default='Filepath')
        parser.add_argument('--asari_command')
        parser.add_argument('-tm', '--table_moniker')
        parser.add_argument('-em', '--empCpd_moniker')
        parser.add_argument('-nm', '--new_moniker')
        parser.add_argument('-cb', '--color_by', default=[])
        parser.add_argument('-bb', '--by_batch')
        parser.add_argument('-mb', '--marker_by', default=[])
        parser.add_argument('-tb', '--text_by', default=[])
        parser.add_argument('--all')
        parser.add_argument('--pca')
        parser.add_argument('--tsne')
        parser.add_argument('--spearman'),
        parser.add_argument('--kendall'),
        parser.add_argument('--missing_feature_distribution')
        parser.add_argument('--missing_feature_percentiles')
        parser.add_argument('--median_correlation_outlier_detection')
        parser.add_argument('--missing_feature_outlier_detection')
        parser.add_argument('--intensity_analysis')
        parser.add_argument('--feature_distribution')
        parser.add_argument('--feature_outlier_detection')
        parser.add_argument('--interactive_plots', default=False)
        parser.add_argument('--save_plots', default=False)
        parser.add_argument('--khipu_isotopes')
        parser.add_argument('--khipu_charges')
        parser.add_argument('--khipu_extended_adducts')
        parser.add_argument('--khipu_adducts_pos')
        parser.add_argument('--khipu_adducts_neg')
        parser.add_argument('--khipu_rt_tolerance')
        parser.add_argument('--blank_value')
        parser.add_argument('--sample_value')
        parser.add_argument('--query_field')
        parser.add_argument('--blank_intensity_ratio')
        parser.add_argument('--drop_name')
        parser.add_argument('--drop_field')
        parser.add_argument('--drop_value')
        parser.add_argument('--drop_others')
        parser.add_argument('--qaqc_filter')
        parser.add_argument('--conversion_command')
        parser.add_argument('--preprocessing_config')
        parser.add_argument('--new_csv_path')
        parser.add_argument('--TIC_normalization_percentile')
        parser.add_argument('--normalize_value')
        parser.add_argument('--feature_retention_percentile')
        parser.add_argument('--interpolation_ratio')
        parser.add_argument('--ms2_dir')
        parser.add_argument('--report_config')
        parser.add_argument('--sample_for_ratio')
        parser.add_argument('--deriv_formula')
        parser.add_argument('--msp_files')
        parser.add_argument('--skip_list')
        parser.add_argument('--add_singletons')
        parser.add_argument('--extra_asari', default=None)
        parser.add_argument('--targets')
        parser.add_argument('--annot_rt_tolerance')
        parser.add_argument('--annot_mz_tolerance')
        parser.add_argument('--accept_licenses')
        parser.add_argument('--scan_experiment', default=False)
        parser.add_argument('--force', default=False)
        parser.add_argument('--file_mode', default="link")
        parser.add_argument('--comprehensive_output', default=False)

        args = parser.parse_args()
        if args.parameters:
            with open(args.parameters, encoding='utf-8') as param_fh:
                params.update(json.load(param_fh))

        for k, v in args.__dict__.items():
            if v:
                params[k] = v
        params['multicores'] = min(mp.cpu_count(), params['multicores'])

        if 'targets' in params:
            if isinstance(params['targets'], str):
                params['targets'] = params['targets'].split()

        for k,v in params.items():
            if isinstance(v, str) and v.endswith(".json"):
                with open(v, encoding='utf-8') as json_fh:
                    params[k] = json.load(json_fh)
        if 'input' in params and not params['input'].endswith("experiment.json"):
            params['input'] = os.path.join(os.path.abspath(params['input']), "experiment.json")
        return params

    @staticmethod
    def download_extras(params):
        """
        This method will download the MoNA LC MS/MS library, and the HMDBv5
        and LMSD in a JMS-compliant format. Currently this downloads from my 
        google drive (I know not ideal). Will be fixed in the future. 

        By using this method you agree to the terms and conditions laid 
        forth in the licenses for each of those repositories

        :param params: This is the master configuration file generated by 
                    parsing the command line arguments plus the defaults. 
        """

        warning = '''
        Pcpfm extras are not actively maintained by the developers of pcpfm and are redistributed forms of third party
        publically available tools. Any issues encountered with these extras may or may not be a problem of pcpfm; 
        however, feel free to raise an issue for us to evaluate. 

        These extras include a JMS-compliant version of the HMDB, LMSD, MoNA.msp files, and the ThermoRawFileParser. 

        All use of these extras are subject to the terms and conditions outlined by their owners. Notably, the HMDB is 
        NOT available for commercial use without a license so please do not use this verison of it for commercial use. 

        Additionally, please cite the original publications for these tools if you use them in your project:

        HMDB - https://hmdb.ca/

        Please cite:
            Wishart DS, Tzur D, Knox C, et al., HMDB: the Human Metabolome Database. Nucleic Acids Res. 2007 Jan;35(Database issue):D521-6. 17202168
            Wishart DS, Knox C, Guo AC, et al., HMDB: a knowledgebase for the human metabolome. Nucleic Acids Res. 2009 37(Database issue):D603-610. 18953024
            Wishart DS, Jewison T, Guo AC, Wilson M, Knox C, et al., HMDB 3.0 — The Human Metabolome Database in 2013. Nucleic Acids Res. 2013. Jan 1;41(D1):D801-7. 23161693
            Wishart DS, Feunang YD, Marcu A, Guo AC, Liang K, et al., HMDB 4.0 — The Human Metabolome Database for 2018. Nucleic Acids Res. 2018. Jan 4;46(D1):D608-17. 29140435
            Wishart DS, Guo AC, Oler E, et al., HMDB 5.0: the Human Metabolome Database for 2022. Nucleic Acids Res. 2022. Jan 7;50(D1):D622–31. 34986597 

        LMSD - https://www.lipidmaps.org/databases/lmsd/overview
        Please cite:
            LMSD: LIPID MAPS® structure databas, Sud M, Fahy E, Cotter D, Brown A, Dennis EA, Glass CK, Merrill AH Jr, Murphy RC, Raetz CR, Russell DW, Subramaniam S., Nucleic Acids Research, 2007, 35: p. D527-32., DOI: 10.1093/nar/gkl838 , PMID: 17098933
            LIPID MAPS® online tools for lipid research, Fahy E, Sud M, Cotter D & Subramaniam S., Nucleic Acids Research, 2007, 35: p. W606-12., DOI: 10.1093/nar/gkm324 , PMID: 17584797
            LIPID MAPS: update to databases and tools for the lipidomics community, Conroy MJ, Andrews RM, Andrews S, Cockayne L, Dennis, EA, Fahy E, Gaud C, Griffiths WJ, Jukes G, Kolchin M, Mendivelso K, Lopez-Clavijo AF, Ready C, Subramaniam S, O'Donnell, VB, Nucleic Acids Research, 2023, DOI: 10.1093/nar/gkad896 , PMID: 37855672 
        
        MoNA - https://mona.fiehnlab.ucdavis.edu/
        Plase cite:
            https://mona.fiehnlab.ucdavis.edu/

        ThermoRawFileParser - 
            Please cite: Niels Hulstaert, Jim Shofstahl, Timo Sachsenberg, Mathias Walzer, Harald Barsnes, Lennart Martens, and Yasset Perez-Riverol Journal of Proteome Research 2020 19 (1), 537-542 DOI: 10.1021/acs.jproteome.9b00328 
        
        By downloading these extras you agree to the terms of their licenses. 

        '''
        print(warning)

        accepted_license = False
        if params['accept_licenses'] in ["True", "TRUE", "Yes", "yes", "T", "Y", "YES", True]:
            accepted_license = True
        else:
            user_input = input("Do you accept their license terms (type yes or no)? ")
            if user_input in ["True", "TRUE", "Yes", "yes", "T", "Y", "YES", True]:
                accepted_license = True

        if accepted_license:
            def download_from_cloud_storage(src, dst, extract_dir=None, delete_after_extract=True, use_gdown=False):
                """
                Downloads a file from either Google Drive or a direct URL, extracts it, and optionally deletes the archive.

                Parameters:
                - src (str): URL or path to the source file in the cloud.
                - dst (str): Local path where the file should be downloaded.
                - extract_dir (str): Directory where the contents should be extracted. 
                                    Defaults to the same directory as dst.
                - delete_after_extract (bool): Whether to delete the archive file after extraction. 
                                            Defaults to True.
                - use_gdown (bool): If True, use gdown to download from Google Drive; 
                                    otherwise, download from a direct URL.
                """
                # Download the file
                if use_gdown:
                    gdown.download(src, output=dst, quiet=False)
                else:
                    response = requests.get(src, stream=True)
                    with open(dst, 'wb') as file:
                        for chunk in response.iter_content(chunk_size=8192):
                            file.write(chunk)
                
                # Determine the extraction directory
                if extract_dir is None:
                    extract_dir = os.path.dirname(dst)

                # Extract the file based on its extension
                if dst.endswith('.zip'):
                    with zipfile.ZipFile(dst, 'r') as zip_ref:
                        zip_ref.extractall(extract_dir)
                elif dst.endswith(('.tar.gz', '.tgz', '.tar')):
                    with tarfile.open(dst, 'r:*') as tar_ref:
                        tar_ref.extractall(extract_dir, filter='data')
                else:
                    raise ValueError(f"Unsupported file format: {dst}")

                # Optionally delete the archive file after extraction
                if delete_after_extract:
                    os.remove(dst)


            this_dir = os.path.abspath(os.path.dirname(__file__))
            thermo_parser_path = os.path.join(this_dir, "ThermoRawFileParser", "ThermoRawFileParser.zip")
            anno_src_path = os.path.join(this_dir, "annotation_sources", "annotation_sources.zip")
            thermo_parser_path = os.path.join(this_dir, "ThermoRawFileParser.zip")
            anno_src_path = os.path.join(this_dir, "annotation_sources.zip")

            download_from_cloud_storage('https://github.com/compomics/ThermoRawFileParser/releases/download/v1.4.4/ThermoRawFileParser1.4.4.zip', thermo_parser_path)
            download_from_cloud_storage('https://storage.googleapis.com/pcpfm-data/annotation_sources-20240119T131612Z-001.zip', anno_src_path, use_gdown=True)

    @staticmethod
    def preprocess(params):
        """
        Using the mappings in the preprocessing config, this will alter
        a provided sequence file and add the extra fields. 

        :param params: This is the master configuration file generated by 
                    parsing the command line arguments plus the defaults. 
        """
        preprocess_config = params['preprocessing_config']
        params["sequence"] = os.path.abspath(params["sequence"])
        with open(params['new_csv_path'], 'w+', encoding='utf-8') as out_csv_fh:
            with open(params['sequence'], encoding='utf-8') as sequence_fh:
                for x, entry in enumerate(csv.DictReader(sequence_fh)):
                    for new_field, _d in preprocess_config["mappings"].items():
                        entry[new_field] = []
                        for new_value, _dd in _d.items():
                            found = False
                            for substring in _dd["substrings"]:
                                for field_to_search in _dd["search"]:
                                    if substring in entry[field_to_search] and found is False:
                                        found = True
                                        entry[new_field].append(new_value)
                            if found is False:
                                if "else" in _dd:
                                    entry[new_field].append(_dd["else"])
                        entry[new_field] = "_".join(entry[new_field])
                    if os.path.exists(params["path_field"]):
                        pass
                    else:
                        sequence_dir = os.path.dirname(params['sequence'])
                        mzml_name = entry[params["name_field"]] + ".mzML"
                        raw_name = entry[params["name_field"]] + ".raw"
                        if os.path.exists(os.path.join(sequence_dir, mzml_name)):
                            entry["InferredPath"] = os.path.join(sequence_dir, mzml_name)
                        elif os.path.exists(os.path.join(sequence_dir, raw_name)):
                            entry["InferredPath"] = os.path.join(sequence_dir, raw_name)
                    if x == 0:
                        writer = csv.DictWriter(out_csv_fh, fieldnames=entry.keys())
                        writer.writeheader()
                    writer.writerow(entry)

    @staticmethod
    def assemble_study(params):
        raise NotImplementedError


    @staticmethod
    def assemble(params):
        """
        This is the first command in any pcpfm analysis. Starting with a 
        sequence file, specified by '-s', an output directory by '-o'
        and a project name specified by '-j', this will create the 
        experiment directory and initialize the experiment.json. 

        Additional arguments include the ability to add a filter on 
        sequence file entries using the '--filter' option and a JSON
        dictionaries.

        <<TODO>>

        Additionally, the --name_field, and --path_field options will 
        allow the user to specify what field name should be used for the
        name and filepath of the acquisitions. Also using --skip_list and 
        a .txt formatted file containing sample_names to ignore, entries 
        can be excluded from an analysis. 
        
        :param params: This is the master configuration file generated by 
                    parsing the command line arguments plus the defaults. 
        """
        experiment = Experiment.Experiment.construct_experiment_from_CSV(
            os.path.join(os.path.abspath(params['output']), str(params['project'])),
            params['sequence'],
            sample_filter=params['filter'],
            name_field=params['name_field'],
            path_field=params['path_field'],
            sample_skip_list_fp=params['skip_list'],
            file_mode=params['file_mode']
        )
        experiment.save()

    @staticmethod
    def convert(params):
        """
        This will convert all .raw files to .mzML using a specified 
        command. To provide the command, you can either modify the 
        config file OR pass the command using the --conversion_command. 
        for this use case, use whatever command will do the conversion but 
        where the .raw file path would be, substitute with $RAW_PATH and 
        where the output would go, put $OUT_PATH. 

        This requires passing -i with the experiment's path.

        :param params: This is the master configuration file generated by 
                    parsing the command line arguments plus the defaults. 
        """
        experiment = Experiment.Experiment.load(params['input'])
        experiment.convert_raw_to_mzML(params['conversion_command'], num_cores=params['multicores'])
        experiment.save()

    @staticmethod
    def asari(params):
        """
        Perform asari on the experiment's acquisitions. They must be have
        been converted or provided in .mzML format first. 

        The command by default assumes a ppm of 5 and the ionization mode 
        of the experiment will be automatically inferred. If extra 
        arguments are desired for asari, they can be provided using
        --extra_asari on the command line. 

        This requires passing -i with the experiment's path.

        :param params: This is the master configuration file generated by 
                    parsing the command line arguments plus the defaults. 
        """
        experiment = Experiment.Experiment.load(params['input'])
        asari_command = params['asari_command']
        if params['extra_asari']:
            asari_command.extend(params['extra_asari'].split(" "))
        experiment.asari(asari_command)
        experiment.save()

    @staticmethod
    def QAQC(params):
        """
        This will perform various QAQC metrics on the indicated feature
        table. By default "all" QAQC metrics are performed which are 
        detailed in the feature table object. 

        This requires passing -i with the experiment's path. 
        The feature table on which to perform the procedures must be given
        as well using either --table_moniker or -tm.

        TODO: this will be deprecated in the future and performed on lazily
        either during report generation or qa/qc filtering. 

        The fields --color_by, --text_by, --marker_by can specify how
        to generate the figures this method generates. For each of these 
        commands, a JSON-formatted list of sequence file fields on which
        to generate the corresponding cosmetic item. These are optional. 

        :param params: This is the master configuration file generated by 
                    parsing the command line arguments plus the defaults. 
        """
        experiment = Experiment.Experiment.load(params['input'])
        experiment.parameters = params["experiment_config"]
        feature_table = experiment.retrieve_feature_table(params['table_moniker'], True)
        if params['table_moniker'] not in experiment.qcqa_results:     
            experiment.qcqa_results[params['table_moniker']] = {}
        for qaqc_result in feature_table.QAQC(params):
            experiment.qcqa_results[params['table_moniker']][qaqc_result["Type"]] = qaqc_result
        experiment.save()

    @staticmethod
    def summarize(params):
        """
        Print the list of empirical compounds and feature tables registered
        wiht the experiment object.

        This requires passing -i with the experiment's path. 

        :param params: This is the master configuration file generated by 
                    parsing the command line arguments plus the defaults. 
        """
        experiment = Experiment.Experiment.load(params['input'])
        experiment.summarize()

    @staticmethod
    def build_empCpds(params):
        """
        For a given feature table, generate empirical compounds from its 
        features. This uses a user-defined set of isotopes and adducts. 

        These can be overwritten, along with other parameters using the 
        follwoing options:

        - --khipu_isotopes specifies the isotopes to use
        - --khipu_adducts specifies which adducts to use
        - --khipu_extended_adducts specifies which extended adducts to use
        - --khipu_adducts_neg specifies which adducts to use if mode is neg
        - --khipu_adducts_pos specifies which adducts to use if mode is pos
        - --add_singletons specifies if we should include single features in \
        the empCpds, i.e., just one peak.
        - --khipu_rt_tolerance the rtime range for which to build khipus
        - --ppm, the mass tolerance for which to build khipus
        - --khipu_charges specifies which charges to consider (absolute Z)

        For details on these parameters, please see Khipu's documentation

        - This requires passing -i with the experiment's path. 
        - This requires passing -tm with the moniker of feature table
        - This requires passing -em with the desired empcpd moniker

        :param params: This is the master configuration file generated by 
                    parsing the command line arguments plus the defaults. 
        """
        experiment = Experiment.Experiment.load(params['input'])
        if experiment.ionization_mode == "pos":
            params['khipu_adducts'] = params['khipu_adducts_pos']
        else:
            params['khipu_adducts'] = params['khipu_adducts_neg']
        EmpCpds.EmpCpds.construct_from_feature_table(experiment,
                                                    params['khipu_isotopes'],
                                                    params['khipu_adducts'],
                                                    params['khipu_extended_adducts'],
                                                    params['table_moniker'],
                                                    params['empCpd_moniker'],
                                                    params['add_singletons'],
                                                    params['khipu_rt_tolerance'],
                                                    params['ppm'],
                                                    params['khipu_charges'])
        exit()
        experiment.save()

    @staticmethod
    def blank_masking(params):
        """
        Print the list of empirical compounds and feature tables registered
        wiht the experiment object.

        - This requires passing -i with the experiment's path. 
        - This requires passing -tm with the feature table's moniker.

        :param params: This is the master configuration file generated by 
                    parsing the command line arguments plus the defaults. 
        """
        experiment = Experiment.Experiment.load(params['input'])
        feature_table = experiment.retrieve_feature_table(params['table_moniker'], True)
        feature_table.blank_mask(params['blank_value'],
                                    params['sample_value'],
                                    params['query_field'],
                                    float(params['blank_intensity_ratio']),
                                    params['by_batch'],
                                    params['batch_blanking_logic'])
        feature_table.save(params['new_moniker'])

    @staticmethod
    def drop_outliers(params):
        """
        This method drop samples from a feature table using the filter in the autodrop json.

        By default this is a |Z| > 2.5 filter on the number of features. This Z-score is calculated
        using the median.

        - This requires passing -i with the experiment's path. 
        - This requires passing -tm with the feature table's moniker.

        :param params: This is the master configuration file generated by 
                    parsing the command line arguments plus the defaults. 
        """
        experiment = Experiment.Experiment.load(params['input'])
        feature_table = experiment.retrieve_feature_table(params['table_moniker'], True)
        feature_table.drop_samples_by_qaqc(params['auto_drop'], False, params=params)
        feature_table.save(params['new_moniker'])

    @staticmethod
    def drop_samples(params):
        """
        This method drop samples from a feature table. There are 
        different modes to use this command in. 

        - --drop_name will drop a sample with a given name
        - --filter will drop samples using a JSON formatted filter
        - --qaqc_filter drops samples using a JSON filter based on qaqc \
            filters
        - --drop_field + --drop_value will drop all samples with a given \
        value for a given field in the sequence file.

        Optionally each command can be augmented by passing the option \
        --drop_others which will reverse the logic of the drop.

        - This requires passing -i with the experiment's path. 
        - This requires passing -tm with the feature table's moniker.

        :param params: This is the master configuration file generated by 
                    parsing the command line arguments plus the defaults. 
        """
        experiment = Experiment.Experiment.load(params['input'])
        feature_table = experiment.retrieve_feature_table(params['table_moniker'], True)
        if params['drop_name']:
            feature_table.drop_sample_by_name(params['drop_name'], params['drop_others'])
        elif params['filter']:
            feature_table.drop_samples_by_filter(params['filter'], params['drop_others'])
        elif params['qaqc_filter']:
            feature_table.drop_samples_by_qaqc(params['qaqc_filter'], params['drop_others'], params=params)
        elif params['drop_field'] and params['drop_value']:
            feature_table.drop_samples_by_field(params['drop_value'], params['drop_field'], params['drop_others'])
        feature_table.save(params['new_moniker'])

    @staticmethod
    def finish(params):
        """
        This command is a no-op command for marking the end of an 
        anlysis in the command history. 

        This requires passing -i with the experiment's path. 

        :param params: This is the master configuration file generated by 
                    parsing the command line arguments plus the defaults. 
        """
        experiment = Experiment.Experiment.load(params['input'])
        experiment.save()

    @staticmethod
    def normalize(params):
        """
        Normalize a feature table based on the TIC of the features 
        present in over a certain percentile of samples. 

        - --TIC_normalization_percentile defines this cutoff
        - --by_batch designates the field to group into batches, if 
           provided, normalization will be done within batches first
        - --normalize_value can be 'mean' or 'median', this will be the 
           value to which the TICs will be normalized

        - This requires passing -i with the experiment's path. 
        - This requires passing -tm with the feature table's moniker.
        - This requires passing -nm with the new feature table's moniker

        :param params: This is the master configuration file generated by 
                    parsing the command line arguments plus the defaults. 
        """
        experiment = Experiment.Experiment.load(params['input'])
        feature_table = experiment.retrieve_feature_table(params['table_moniker'], True)
        if params["TIC_normalization_percentile"]:
            feature_table.TIC_normalize(float(params["TIC_normalization_percentile"]),
                                        params["by_batch"],
                                        params["normalize_value"])
        feature_table.save(params['new_moniker'])

    @staticmethod
    def drop_missing_features(params):
        """
        Drop samples below a given percentile of inclusion. 

        - --feature_retention_percentile defines this cutoff
        - --by_batch designates the field to group into batches, if 
           provided, the percentile is caluclated per batch first
        - --feature_drop_logic can be "or" or "and" and specifies how 
           handle the various batches. For example, if "or", a feature
           will be dropped if it is below the cutoff in any batch.

        - This requires passing -i with the experiment's path. 
        - This requires passing -tm with the feature table's moniker.
        - This requires passing -nm with the new feature table's moniker.

        :param params: This is the master configuration file generated by 
                    parsing the command line arguments plus the defaults. 
        """
        experiment = Experiment.Experiment.load(params['input'])
        feature_table = experiment.retrieve_feature_table(params['table_moniker'], True)
        feature_table.drop_missing_features(params["by_batch"],
                                            float(params["feature_retention_percentile"]),
                                            params["feature_drop_logic"])
        feature_table.save(params['new_moniker'])

    @staticmethod
    def impute(params):
        """
        Replace remaining missing values with a value to aid statistics
        
        - --interpolation_ratio this value specifies what to multiply the 
          value generated by the interpolate_method before replacement
        - --interpolate_method currently limited to only min
        - --by_batch this field specifies what field to group samples by 
          and interpolates within each group (probably a bad idea)

        - This requires passing -i with the experiment's path. 
        - This requires passing -tm with the feature table's moniker.
        - This requires passing -nm with the new feature table's moniker.

        :param params: This is the master configuration file generated by 
                    parsing the command line arguments plus the defaults. 
        """
        experiment = Experiment.Experiment.load(params['input'])
        feature_table = experiment.retrieve_feature_table(params['table_moniker'], True)
        feature_table.impute_missing_features(float(params['interpolation_ratio']),
                                                    params['by_batch'],
                                                    params['interpolate_method'])
        feature_table.save(params['new_moniker'])

    @staticmethod
    def batch_correct(params):
        """
        Use pyCombat to correct for batch effects using the specified batch
        identifier.

        - This requires passing -i with the experiment's path. 
        - This requires passing -tm with the feature table's moniker.
        - This requires passing -nm with the new feature table's moniker.
        - This requires passing --by_batch with the field specifying the 
            batch on which to correct

        :param params: This is the master configuration file generated by 
                    parsing the command line arguments plus the defaults. 
        """
        experiment = Experiment.Experiment.load(params['input'])
        feature_table = experiment.retrieve_feature_table(params['table_moniker'], True)
        feature_table.batch_correct(params['by_batch'])
        feature_table.save(params['new_moniker'])

    @staticmethod
    def delete(params):
        """
        Delete a specified feature table or empCpd list by moniker.

        - This requires passing -i with the experiment's path. 
        - This requires passing -tm with the table's moniker to delete
                                   or
        - This requires passing -em with the empcpd's moniker to delete

        Note: you *cannot* delete the feature tables generated by \
        asari using this method. 

        :param params: This is the master configuration file generated by 
                    parsing the command line arguments plus the defaults. 
        """
        experiment = Experiment.Experiment.load(params['input'])
        if params["table_moniker"]:
            experiment.delete_feature_table(params['table_moniker'])
        elif params['empCpd_moniker']:
            experiment.delete_empCpds(params['empCpd_moniker'])

    @staticmethod
    def log_transform(params):
        """
        Log transform a given table, by default, log2

        --log_transform_mode can be log10 or log2

        - This requires passing -i with the experiment's path. 
        - This requires passing -tm with the table's moniker to transform
        - This requires passing -nm with the new feature table's moniker

        :param params: This is the master configuration file generated by 
                    parsing the command line arguments plus the defaults. 
        """
        experiment = Experiment.Experiment.load(params['input'])
        feature_table = experiment.retrieve_feature_table(params['table_moniker'], True)
        feature_table.log_transform(params['log_transform_mode'])
        experiment.log_transformed_feature_tables.append(params["new_moniker"])
        feature_table.save(params["new_moniker"])


    @staticmethod
    def l4_annotate(params):
        """
        This will generate MS1 annotations on a provided feature table
        or empcpd list. 

        - **--log_transform_mode**: can be log10 or log2
        - **--targets**: will specify what compounds to annotate, must be a \
                        JMS-compliant JSON file
        - **--annot_mz_tolerance**: this is the ppm cutoff for the search
        - **--annot_rt_tolerance**: this is the rtime cutoff, in sec, for the search

        ..

        - This requires passing -i with the experiment's path. 
        - This requires passing -tm with the table's moniker to annotate or
        - This requires passing -em with the empCpd's moniker to annotate
        - This requires passing -nm with the new moniker for the table or empcpd list.

        :param params: This is the master configuration file generated by 
                    parsing the command line arguments plus the defaults. 
        """

        experiment = Experiment.Experiment.load(params['input'])
        if 'empCpd_moniker' in params:
            empCpd = experiment.retrieve_empCpds(params['empCpd_moniker'], True)
            empCpd.l4_annotate(params['targets'], float(params['annot_rt_tolerance']))
            empCpd.save(params['new_moniker'])

    @staticmethod
    def l2_annotate(params):
        """
        This will generate MS2 annotations on a provided feature table
        or empCpd list. Requires that MS2 spectra first be mapped.

        - **--msp_files**: Designate the path to the MSP files to use for annotation.
        - **--annot_mz_tolerance**: PPM cutoff for the precursor ion search, default = 5ppm.
        - **--annot_rt_tolerance**: Time cutoff, in seconds, for the precursor ion search, default = 30sec.
        - **--ms2_similarity_metric**: Name of any matchms method for comparing MS2 spectra, default = CosineHungarian.
        - **--ms2_min_peak**: Minimum number of matching peaks required for an MS2 match, default = 3.
        
        ..

        - This requires passing -i with the experiment's path. 
        - This requires passing -tm with the table's moniker to annotate or 
        - This requires passing -em with the empCpd's moniker to annotate
        - This requires passing -nm with the new moniker for the table or empcpd list.

        :param params: This is the master configuration file generated by 
                    parsing the command line arguments plus the defaults.
        """
        experiment = Experiment.Experiment.load(params['input'])
        if 'msp_files' in params:
            msp_file = params['msp_files']
            if isinstance(msp_file, str):
                if msp_file.endswith(".json"):
                    with open(msp_file, encoding='utf-8') as msp_fh:
                        msp_file = json.load(msp_fh)
                else:
                    msp_file = [msp_file]
        elif experiment.ionization_mode == "pos":
            msp_file = params['msp_files_pos']
        elif experiment.ionization_mode == "neg":
            msp_file = params['msp_files_neg']
        if 'empCpd_moniker' in params:
            empCpd = experiment.retrieve_empCpds(params['empCpd_moniker'], True)
            empCpd.l2_annotate(
                msp_file,
                params["annot_mz_tolerance"],
                params["ms2_similarity_metric"],
                params["ms2_min_peaks"],
            )
            empCpd.save(params["new_moniker"])

    @staticmethod
    def l1b_annotate(params):
        """
        This will generate level 1 annotations on a empcpd list using
        a csv file(s) with compound names, retention times and m/z 
        values. 

        - **--targets**: a list of csv filepaths with mz, retention times, compound \
            names with column names, "mz", "R", "CompoundName"
        - **--annot_mz_tolerance**: the ppm cutoff for the precursor ion search, default = 5 ppm
        - **--annot_rt_tolerance**: the rtime cutoff, in sec, for the precursor ion search, default = 30 sec

        ..

        - This requires passing -i with the experiment's path. 
        - This requires passing -em with the empCpd's moniker to annotate 
        - This requires passing -nm with the new moniker for the empcpd list.

        :param params: This is the master configuration file generated by parsing the command line arguments plus the defaults.
        """
        experiment = Experiment.Experiment.load(params['input'])
        if 'empCpd_moniker' in params:
            empCpd = experiment.retrieve_empCpds(params['empCpd_moniker'], True)
            empCpd.l1b_annotate(params['targets'],
                                float(params['annot_rt_tolerance']), 
                                float(params['annot_mz_tolerance'])
                                )
            empCpd.save(params['new_moniker'])

    @staticmethod
    def l1a_annotate(params):
        """
        This will generate level 1 annotations on a empcpd list using
        a csv file(s) with compound names, retention times and m/z 
        values. 

        --targets are a list of csv filepaths with mz, retention 
        times, compound names with column names, "mz", "rtime",
        "CompoundName"
        --annot_mz_tolerance this is the ppm cutoff for the precursor 
            ion search
        --annot_rt_tolerance this is the rtime cutoff, in sec, for
            the precursor ion search

        This requires passing -em with the empCpd's moniker to annotate
        This requires passing -nm with the new moniker for the table or 
            empcpd list.

        :param params: This is the master configuration file generated by 
        parsing the command line arguments plus the defaults. 
        """
        experiment = Experiment.Experiment.load(params['input'])
        if 'empCpd_moniker' in params:
            empCpd = experiment.retrieve_empCpds(params['empCpd_moniker'], True)
            empCpd.l1a_annotate(params['targets'],
                                float(params['annot_rt_tolerance']), 
                                float(params['annot_mz_tolerance']),
                                )
            empCpd.save(params['new_moniker'])

    @staticmethod
    def report(params):
        """
        This will generate a pdf report using a JSON template 

        - **--report_config** will override the default template

        - This requires passing -i with the experiment's path. 

        :param params: This is the master configuration file generated by 
                    parsing the command line arguments plus the defaults. 
        """
        experiment = Experiment.Experiment.load(params['input'])
        Report.Report(experiment, params)

    @staticmethod
    def map_ms2(params):
        """
        This maps MS2 spectra to the empCompounds based on rt and mz similarity. 

        Once mapped, they can be annotated using MS2 similarity via l2_annotate and l1a_annotate.

        --annot_mz_tolerance this is the ppm cutoff for the precursor 
            ion search, default is 5 ppm
        --annot_rt_tolerance this is the rtime cutoff, in sec, for
            the precursor ion search, default is 30 sec

        This will scan for all MS2 spectra in the experiment. Additional MS2, from AcquireX for 
        example, can be added by specifying the path to them using --ms2_dir.
        
        - This requires passing -i with the experiment's path. 
        - This requires passing -em with the empCpd's moniker to annotate
        - This requires passing -nm with the new moniker for the empcpd list.

        :param params: This is the master configuration file generated by 
                    parsing the command line arguments plus the defaults. 
        """
        experiment = Experiment.Experiment.load(params['input'])
        empCpds = experiment.retrieve_empCpds(params['empCpd_moniker'], True)
        empCpds.map_ms2(float(params['annot_rt_tolerance']), 
                        float(params['annot_mz_tolerance']), 
                        ms2_files=params['ms2_dir'],
                        scan_experiment=params['scan_experiment'])
        empCpds.save(params["new_moniker"])

    @staticmethod
    def generate_output(params):
        """
        This command generates the three table output for downstream
        analysis. This includes a feature table, an annotation table,
        and finally the sample metadata. All results are stored in the
        results subdirectory according to the specified moniker. 

        - This requires passing -i with the experiment's path.
        - This requires passing -tm for the table moniker to include
        - This requires passing -em for the empcpd moniker to include
        - This requires passing -nm for the new moniker to save 
            generated results using. 

        If you want to have all the feature information in the annotation
        table, please add the following:

        --comprehensive_output true

        :param params: This is the master configuration file generated by 
                    parsing the command line arguments plus the defaults. 
        """
        experiment = Experiment.Experiment.load(params['input'])
        experiment.generate_output(params['empCpd_moniker'], params['table_moniker'], params['comprehensive_output'])

    @staticmethod
    def reset(params):
        """
        This command resets the experiment object back to when asari was
        ran. This removes all user-generated monikered entities, including
        feature tables and empirical compounds. This removes any qcqa results
        that were generated along with any qaqc figures. 

        This is useful if you want to restart an analysis but you do not want
        to rerun asari, reconvert the acquisitions, or recreate the experiment. 

        This action is irreversible

            - This requires passing -i with the experiment's path.
            - If passed, --force, will skip the user confirmation 
        """

        warning = """
        THIS WILL DELETE ALL USER-GENERATED MONIKERED ENTITIES
        
        THIS INCLUDES: 
            emp_cpds
            feature_tables

        AND ANY QAQC RESULTS / FIGURES GENERATED FOR THESE ENTITES

        INPUT 'yes' TO PROCEED. ALTERNATIVELY PROVIDE '--force' TO
        IGNORE THIS WARNING (FOR HANDS-OFF USAGE). ANY OTHER INPUT
        WILL CANCEL THE RESET.

        """

        if not params['force']:
            print(warning)
            consent = input()
            if consent not in {'yes', 'YES', 'y', 'Y'}:
                print("Cancelling reset!")
                return
        experiment = Experiment.Experiment.load(params['input'])
        experiment.delete_feature_table('*')
        experiment.delete_empCpds('*')
        experiment.qaqc_figs = None
        experiment.qcqa_results = None
        experiment.log_transformed_feature_tables = None
        experiment.cosmetics = None
        experiment.used_cosmetics = None
        experiment.save()

def main():
    """
    This is the main function for the pipeline
    """
    params = Main.process_params()

    print("Attempting: ", params["subcommand"])
    if params['subcommand'] not in dir(Main):
        print(params['subcommand'] + " is not a valid subcommand")
        print("valid commands include:")
        for method in dir(Main):
            if not method.startswith('__'):
                print("\t", method)
    else:
        function = getattr(Main, params['subcommand'])
        try:
            function(params)
            print("Succesfully executed: ", params["subcommand"])
        except Exception as e:
            import traceback
            print("Error executing: " + params['subcommand'])
            print(function.__doc__)
            print("Exception: ", e)
            print("Traceback:")
            print(traceback.format_exc())

def CLI():
    '''
    This function is called when 'pcpfm' is called in the terminal. 

    Simply a wrapper around main()
    '''
    main()

if __name__ == '__main__':        
    main()


'''
Input/Output functions

>>> import sys
>>> sys.path.append('jms')
>>> from jms.io import read_hmdb_to_empCpds
>>> f = '/Users/shuzhao/li.github/mass2chem/mass2chem/source_data/HMDB4_compounds.tsv'
>>> list_empCpds = read_hmdb_to_empCpds(f)
>>> list_empCpds[5559]
{'interim_id': 'C13H18N4O6_326.122634', 'neutral_formula': 'C13H18N4O6', 'neutral_formula_mass': 326.122634328, 
'compounds': [{'id': 'HMDB0003826', 'name': '6,7-Dimethyl-8-(1-D-ribityl)lumazine'}]}
>>> import json
>>> with open('jms/data/compounds/hmdb4_compounds.json', 'w', encoding='utf-8') as f:
...   json.dump(list_empCpds, f, ensure_ascii=False, indent=2)
... 
>>> len(list_empCpds)
11937

# example of a Compound:
        {'primary_id': 'HMDB0042472', 'primary_db': 'HMDB', 
        'name': 'TG(14:0/22:1(13Z)/22:4(7Z,10Z,13Z,16Z))', 
        'neutral_formula': 'C61H108O6', 'neutral_formula_mass': 936.814591198, 
        'SMILES': '[H][C@](COC(=O)CCCCCCCCCCCCC)(COC(=O)CCCCC\\C=C/C\\C=C/C\\C=C/C\\C=C/CCCCC)OC(=O)CCCCCCCCCCC\\C=C/CCCCCCCC', 
        'inchikey': 'NDYSJNXSSRKHCQ-LHDKJNFNSA-N', 
        'other_ids': {'PubChem': '131753609', 'KEGG': '', 'ChEBI': ''}}
'''

peak_attribute_dict = {
    'm/z': 'mz',
    'mz': 'mz',
    'apex': 'rtime',
    # ...,
}

def read_table_to_peaks(infile, 
                        has_header=True, mz_col=1, rtime_col=2, intensity=None, feature_id=None,
                        full_extract=True, max_col=None,
                        delimiter='\t'):
    '''
    Read a text feature table, and 
    return list of peaks, e.g. [ {
        'id_number': 555,
        'mz': 133.0970, 
        'apex': 654, 
        'height': 14388.0, 
        }, ... ]
    
    full_extract: to keep all fields in output as strings, only if has_header.
    feature_id: if None, create id for each peak/feature.
    '''
    def _make_id(ii, mz, rt):
        return 'F' + str(ii) + '_' + str(round(mz, 6)) + '@' + str(round(rt, 2))

    with open(infile) as ft_fh:
        ft = ft_fh.readlines()
    if has_header:
        header = ft[0].rstrip().split(delimiter)
        if max_col and not full_extract:
            header = header[:max_col]
        ft = ft[1:]
    list_peaks = []
    for ii, entry in enumerate(ft):
        values = entry.split(delimiter)
        if max_col and not full_extract:
            values = line[:max_col]
        mz, rt = float(values[mz_col]), float(values[rtime_col])
        fid = values[feature_id].strip() if feature_id is not None else _make_id(ii, mz, rt)
        representative_intensity = [float(x) for x in values[intensity[0]: intensity[1]]] if intensity else 0
        peak = {'id_number': fid, 'id': fid, 'mz': mz, 'rtime': rt, 'apex': rt, 'representative_intensity': representative_intensity}
        if has_header and full_extract:
            for k, v in zip(header, values):
                if k not in peak:
                    peak[k] = float(v)
        list_peaks.append(peak)
    return list_peaks



def read_tsv_hmdb_to_empCpds(infile, delimiter='\t'):
    '''
    Previously parsed HMDB4 in short tsv format.
    Ref:
    https://github.com/shuzhao-li/JMS/wiki/Scripting-to-create-a-dictionary-from-HMDB-data
    https://github.com/shuzhao-li/JMS/blob/main/jms/utils/hmdb.py
    A copy of HMDB4_compounds.tsv is under
    https://github.com/shuzhao-li/mass2chem/tree/master/mass2chem/lib

    In [26]: hmdb[0].split('\t')                                                                                              
    Out[26]: 
    ['accession',
    'name',
    'chemical_formula',
    'monisotopic_molecular_weight',
    'kegg_id',
    'pubchem_compound_id\n']

    This is not producing exact format as in dbStructures, which should be followed when needed more elaborate.
    '''
    hmdb = open(infile).readlines()
    hdict = {}                                                                                                       
    for line in hmdb[1:]: 
        a = line.split(delimiter) 
        if a[3].strip(): 
            hdict[a[0]] = a 
    h0 = [x.split('\t')[0] for x in hmdb[1:]] 
    massDict_hmdb = {}
    for x in h0:  
        try: 
            a = hdict[x]  
            k = a[2]+ '_' + str(round(float(a[3]),6))  # ensuring identical formula and mass
            if k in massDict_hmdb:  
                massDict_hmdb[k].append( a )  
            else:  
                massDict_hmdb[k] = [a]  
        except KeyError: 
            print(x) 
    list_empCpds = []
    for k,v in massDict_hmdb.items():
        compounds = []
        for cpd in v:
            compounds.append({
                "id": cpd[0], "name": cpd[1]
            })
        list_empCpds.append(
            {
                "interim_id": k,
                "neutral_formula": v[0][2],
                "neutral_formula_mass": float(v[0][3]),
                "compounds": compounds,
            }
        )
    return list_empCpds





'''
This module is concerned with the construction of EmpCpds and their annotation.
'''

import json
import os
import pandas as pd

from jms.io import read_table_to_peaks
from jms.dbStructures import ExperimentalEcpdDatabase, knownCompoundDatabase
from intervaltree import IntervalTree
from khipu.epdsConstructor import epdsConstructor
from khipu.extended import (
    isotope_search_patterns,
    extended_adducts,
    adduct_search_patterns,
    adduct_search_patterns_neg,
)
from .MSnSpectrum import MS2Spectrum
from .utils import (
    get_similarity_method,
    lazy_extract_ms2_spectra,
    search_for_mzml,
    extract_CD_csv,
)

class EmpCpds:
    """
    This object is largely a warpper around the dict_empcpds returned from Khipu.
    """
    def __init__(self, dict_empcpds, experiment, moniker):
        """
        the empCpds object is a wrapper around dict_empcpds that will associate the dict_empcpds
        with a moniker and experiment object.

        Args:

        dict_empcpds (dict): dictionary of empCpds, from khipu
        experiment (object): experiment object for these empdpds
        moniker (str): the moniker for this empCpd
        """
        self.dict_empcpds = dict_empcpds
        self.experiment = experiment
        self.moniker = moniker

        self._feature_id_to_khipu_id = None
        self._khipu_id_to_feature_id = None

        # these are lazily evaluated
        self.__mz_trees = {}
        self.__rt_trees = {}
        self.__ms2_spectra = None

    @property
    def ms2_spectra(self):
        """
        This is a lazily evaluated data store for MS2 spectra

        Returns:
            dict: ms2_id to ms2_spetra dictionary.
        """        
        if self.__ms2_spectra is None:
            ms2_spectra = {}
            for khipu in self.dict_empcpds.values():
                if "MS2_Spectra" in khipu:
                    for spectrum in khipu["MS2_Spectra"]:
                        if spectrum["precursor_ion_id"] not in ms2_spectra:
                            ms2_spectra[spectrum["precursor_ion_id"]] = [MS2Spectrum.from_embedding(spectrum)]
                        else:
                            ms2_spectra[spectrum["precursor_ion_id"]].append(MS2Spectrum.from_embedding(spectrum))
            self.__ms2_spectra = ms2_spectra
        return self.__ms2_spectra

    @property
    def feature_id_to_khipu_id(self):
        """
        This property provides a mapping from feature ids back to the khipu that contains them.

        Returns:
            dict: feature to kp id mapping dict
        """
        if self._feature_id_to_khipu_id is None:
            feature_id_to_khipu_id = {}
            khipu_id_to_feature_id = {}
            for kp_id, khipu in self.dict_empcpds.items():
                khipu_id_to_feature_id[kp_id] = []
                for peak in khipu["MS1_pseudo_Spectra"]:
                    feature_id_to_khipu_id[peak["id_number"]] = kp_id
                    khipu_id_to_feature_id[kp_id].append(peak["id_number"])
            self._feature_id_to_khipu_id = feature_id_to_khipu_id
            self._khipu_id_to_feature_id = khipu_id_to_feature_id
        return self._feature_id_to_khipu_id

    @property
    def khipu_id_to_feature_id(self):
        """
        This property provides a mapping of khipu id to the feature ids in the khipu

        Returns:
            dict: kp_id to the feature ids
        """
        if self._feature_id_to_khipu_id is None:
            feature_id_to_khipu_id = {}
            khipu_id_to_feature_id = {}
            for kp_id, khipu in self.dict_empcpds.items():
                khipu_id_to_feature_id[kp_id] = []
                for peak in khipu["MS1_pseudo_Spectra"]:
                    feature_id_to_khipu_id[peak["id_number"]] = kp_id
                    khipu_id_to_feature_id[kp_id].append(peak["id_number"])
            self._feature_id_to_khipu_id = feature_id_to_khipu_id
            self._khipu_id_to_feature_id = khipu_id_to_feature_id
        return self._khipu_id_to_feature_id

    def create_annotation_table(self, comprehensive_output=False):
        """
        This flattens the empcpd annotations into a dataframe summarizing the annotation on a 
        per-feature level.
        
        This is for the generation of outputs.

        Returns:
            dataframe: annotation table
        """

        #ion_relation_lookup = {}
        #for kp_id, kp in self.dict_empcpds.items():
        #    for ms1_peak in kp['MS1_pseudo_Spectra']:
                
        feature_lookup = {}
        for kp_id, kp in self.dict_empcpds.items():
            for ms1_peak in kp["MS1_pseudo_Spectra"]:
                feature_lookup[ms1_peak['id']] = ms1_peak

        annotation_table = []
        for kp_id, khipu in self.dict_empcpds.items():
            for feature in self.khipu_id_to_feature_id[kp_id]:
                if comprehensive_output:
                    annotation_entry = feature_lookup[feature]
                else:
                    annotation_entry = {}
                #MS1 annots
                l1b_annots = khipu.get("Level_1b", [])
                for l1b_annot in l1b_annots:
                    l1b_annot_entry = {k: v for k, v in annotation_entry.items()}
                    l1b_annot_entry.update({"feature": feature, "level": "1b"})
                    l1b_annot_entry.update({"name": l1b_annot[0], "source": l1b_annot[1]})
                    annotation_table.append(l1b_annot_entry)
                l4_annots = khipu.get("Level_4", [])
                for l4_annot in l4_annots:
                    l4_annot_entry = {k: v for k, v in annotation_entry.items()}
                    l4_annot_entry.update({"feature": feature, "level": "4"})
                    l4_annot_entry.update(l4_annot)
                    annotation_table.append(l4_annot_entry)
                #MS2 annots
                for ms2_spectrum in khipu.get("MS2_Spectra", []):
                    for annotation in ms2_spectrum["annotations"]:
                        annotation_level = annotation["annotation_level"]
                        ms2_annotation = {k: v for k, v in annotation_entry.items()}
                        ms2_annotation = {"feature": feature, "level": annotation_level}
                        ms2_annotation.update(annotation)
                        annotation_table.append(ms2_annotation)
        return pd.DataFrame(annotation_table)
    
    def __update_ms2(self):
        """
        This method will iterate through all the ms2 spectra in the ms2 property and maps them
        back to the actual khipu objects. 

        This is not elegant, but it works. 

        This needs to be called whenever the MS2 annotations are updated before saving the 
        empcpd object.
        """        
        for khipu in self.dict_empcpds.values():
            if "MS2_Spectra" in khipu:
                new_spectra = []
                for spectrum in khipu["MS2_Spectra"]:
                    for match in self.ms2_spectra[spectrum["precursor_ion_id"]]:
                        new_spectra.append(match.embedding())
                khipu["MS2_Spectra"] = new_spectra

    def update_annotations(self, update_ms2=False):
        """
        This method iterates through all khipus and updates the relevant annotation fields. 

        Args:
            update_ms2 (bool): if True, update the MS2 annotations, default is False.
        """
        if update_ms2:
            self.__update_ms2()
        # update other fields in the empCpds:
        for _, khipu in self.dict_empcpds.items():
            khipu["identity"] = set()
            khipu["Database_referred"] = set() 

            # MS1 only, only Level_1b counts for identity
            if "Level_1b" in khipu:
                for annotation, source in khipu["Level_1b"]:
                    khipu["identity"].add((annotation, "no score - ms1 only"))
                    khipu["Database_referred"].add(source)

            # MS1 only, NOT IDENTITY
            if "Level_4" in khipu:
                for annotation in khipu["Level_4"]:
                    khipu["Database_referred"].add(annotation["primary_db"])

            # MS2, both Level_1a and Level 2 will update identity
            if "MS2_Spectra" in khipu:
                for ms2_spectrum in khipu["MS2_Spectra"]:
                    for annotation in ms2_spectrum["annotations"]:
                        if annotation["annotation_level"] == "Level_2":
                            khipu["identity"].add(
                                (
                                    annotation["reference_id"],
                                    round(annotation["msms_score"], 3),
                                )
                            )
                            if "primary_db" in annotation:
                                khipu["Database_referred"].add(annotation["primary_db"])
                            else:
                                khipu["Database_referred"].add("MS2")
                        if annotation["annotation_level"] == "Level_1a":
                            khipu["identity"].add(
                                (
                                    annotation["reference_id"],
                                    round(annotation["msms_score"], 3),
                                    "matches standard",
                                )
                            )
                            if "primary_db" in annotation:
                                khipu["Database_referred"].add(annotation["primary_db"])
                            else:
                                khipu["Database_referred"].add("MS2")
            khipu["identity"] = list(khipu["identity"])
            khipu["Database_referred"] = list(khipu["Database_referred"])

    def get_mz_tree(self, mz_tol, abs_error=False):
        """
        This method will return an existing m/z based interval tree for
        these empcpds for a given mz_tol.

        Args:
            mz_tol (float): the mz_tol assumed to be in ppm
            abs (bool): if true, assume the mz tolerance provide is in daltons

        Returns:
            intervaltree: interval tree for mz at the provided mz_tol
        """
        if ("feature", str(mz_tol), str(abs_error)) not in self.__mz_trees:
            mz_tree = IntervalTree()
            for _, khipu in self.dict_empcpds.items():
                for peak in khipu["MS1_pseudo_Spectra"]:
                    if abs_error:
                        mz_tree.addi(
                            peak["mz"] - mz_tol,
                            peak["mz"] + mz_tol,
                            peak["id_number"],
                        )
                    else:
                        mz_error = peak["mz"] / 1e6 * mz_tol
                        mz_tree.addi(
                            peak["mz"] - mz_error,
                            peak["mz"] + mz_error,
                            peak["id_number"],
                        )
            self.__mz_trees[("feature", str(mz_tol), str(abs_error))] = mz_tree
        return self.__mz_trees[("feature", str(mz_tol), str(abs_error))]

    def get_rt_tree(self, rt_tolerance):
        """
        This method will return an existing rt based interval tree for
        these empcpds for a given rt_tolerance

        Args:
            mz_tol (float): the rt_tolerance in sec(s)

        Returns:
            intervaltree: interval tree for rtime at the provided rt tolerance
        """
        if ("feature", rt_tolerance) not in self.__rt_trees:
            rt_tree = IntervalTree()
            for _, khipu in self.dict_empcpds.items():
                for peak in khipu["MS1_pseudo_Spectra"]:
                    rt_tree.addi(
                        peak["rtime"] - rt_tolerance,
                        peak["rtime"] + rt_tolerance,
                        peak["id_number"],
                    )
            self.__rt_trees[("feature", rt_tolerance)] = rt_tree
        return self.__rt_trees[("feature", rt_tolerance)]

    def get_precursor_mz_tree(self, mz_tol):
        """
        This retrieves or generates the mz tree of all precursor ions for the empCpd MS2 spectra 
        at a given ppm mass tolerance.

        Args:
            mz_tol (float): the mz tolerance in ppm

        Returns:
            intervaltree: an interval tree for all precursor ion mzs at the given mz tolerance.
        """
        if ("precursor", mz_tol) not in self.__mz_trees:
            mz_tree = IntervalTree()
            for spectra in self.ms2_spectra.values():
                for spectrum in spectra:
                    precursor_mz = spectrum.prec_mz
                    mz_error = precursor_mz / 1e6 * mz_tol
                    mz_tree.addi(
                        precursor_mz - mz_error,
                        precursor_mz + mz_error,
                        spectrum.precursor_ion_id,
                    )
            self.__mz_trees[("precursor", mz_tol)] = mz_tree
        return self.__mz_trees[("precursor", mz_tol)]

    def get_precursor_rt_tree(self, rt_tolerance):
        """
        This retrieves or generates the retention time tree of all precursor ions for the empCpd
        MS2 spectra at a given ppm mass tolerance.

        Args:
            rt_tolerance (float): the rtime tolerance in seconds

        Returns:
            intervaltree: an interval tree for all precursor ion rtimes at the given rt tolerance.
        """
        if ("precursor", rt_tolerance) not in self.__rt_trees:
            rt_tree = IntervalTree()
            for spectra in self.ms2_spectra.values():
                for spectrum in spectra:
                    precursor_rt = spectrum.rtime
                    rt_tree.addi(
                        precursor_rt - rt_tolerance,
                        precursor_rt + rt_tolerance,
                        spectrum.precursor_ion_id,
                    )
            self.__mz_trees[("precursor", rt_tolerance)] = rt_tree
        return self.__mz_trees[("precursor", rt_tolerance)]

    @property
    def num_khipus(self):
        """
        This method returns the number of khipus in empCpd

        int: number of empcpds
        """
        return len(self.dict_empcpds)

    @property
    def num_features(self):
        """
        This method returns the number of features contained within
        the empcpds.

        int: number of features in empcpds.

        """
        return len(self.feature_id_to_khipu_id)

    def search_for_feature(
        self, query_mz=None, query_rt=None, mz_tol=None, rt_tolerance=None
    ):
        """
        Given a query_mz and query_rt with corresponding tolerances in ppm and absolute units respectively find all
        features by id_number that have a matching mz and rtime.

        All search fields are optional but if none are provided then all the features will be considered matching.
        The mz tolerance should be in ppm while the rtime tolerance should be provided in rtime units.

        Args:
            
        :param query_mz: the mz to search for, defaults to None
        :type query_mz: float, optional
        :param query_rt: the rtime to search for, defaults to None
        :type query_rt: float, optional
        :param mz_tol: the tolerance in ppm for the mz match, defaults to None
        :type mz_tol: float, optional
        :param rt_tolerance: the tolerance in absolute units for the rt match, defaults to None
        :type rt_tolerance: float, optional
        :return: list of matching feature IDs
        :rtype: list
        """
        mz_matches, rt_matches = set(), set()
        if query_mz and mz_tol:
            if isinstance(mz_tol, str):
                if mz_tol.endswith("ppm"):
                    mz_tol = float(mz_tol.rstrip("ppm"))
                    mz_tree = self.get_mz_tree(mz_tol)
                    mz_matches = {x.data for x in mz_tree.at(query_mz)}
                elif mz_tol.endswith("amu"):
                    mz_tol = float(mz_tol.rstrip("amu"))
                    mz_tree = self.get_mz_tree(mz_tol, abs_error=True)
                    mz_matches = {x.data for x in mz_tree.at(query_mz)}
            else:
                mz_tree = self.get_mz_tree(mz_tol)
                mz_matches = {x.data for x in mz_tree.at(query_mz)}
        if query_rt and rt_tolerance:
            rt_matches = {x.data for x in self.get_rt_tree(rt_tolerance).at(query_rt)}
        return list(rt_matches.intersection(mz_matches))

    def save(self, save_as_moniker=None):
        """
        This method saves the empirical compound dictionary to the annotation_subdirectory.
        The path is determined by the moniker for the empCpds object, however, an alternative moniker can be provided
        which effectively saves a new empCpd object. This also updates the empCpds registry in the experiment with the
        path to the stored json.

        :param save_as_monhiker: an alternative moniker to which to save the table. Defaults to None.
        """
        save_as_moniker = self.moniker if save_as_moniker is None else save_as_moniker
        self.experiment.empCpds[save_as_moniker] = os.path.join(
            self.experiment.annotation_subdirectory, save_as_moniker + "_empCpds.json"
        )
        with open(self.experiment.empCpds[save_as_moniker], "w+", encoding='utf-8') as out_fh:
            json.dump(self.dict_empcpds, out_fh, indent=4)
        self.experiment.save()

    @staticmethod
    def load(moniker, experiment):
        """
        This method generates the empCpd object for the provided moniker.

        :param moniker: the empCpd moniker to load
        :param experiment: the experiment from which the empCpd was
            generated
        :return: the empCpds object for the specified moniker
        """
        with open(experiment.empCpds[moniker], encoding='utf-8') as empcpd_fh:
            return EmpCpds(json.load(empcpd_fh), experiment, moniker)

    def map_ms2(
        self,
        mapping_mz_tol=5,
        mapping_rt_tolerance=30,
        ms2_files=None,
        scan_experiment=False,
    ):
        """
        When MS2 data is acquired, each spectrum will have a retention time and precursor 
        ion mz. These can be mapped to features in the empCpds before annotation thus 
        limiting any subsequent searches to just the MS2 spectra that appear to represent
        features that we care about. 

        By default this method searches all acquisitions in the experiment for MS2 spectra.

        Additional MS2 spectra can be provided as mzml files using the ms2_files param.

        Args:
            mapping_mz_tol (float, optional): mz tolerance for the feature, ion precursor mz 
            match in ppm. Defaults to 5.
            mapping_rt_tolerance (int, optional): rt tolerance for the feature, ion precursor time match 
            in seconds. Defaults to 30.
            ms2_files (str, optional): path to additional ms2 acquisitions. Defaults to None.
            scan_experiment (bool, optional): _description_. Defaults to False.
        """
        # we should first map ms2 spectra to empCpds, then annnotate them.
        mzml_w_ms2 = []
        if ms2_files:
            if ms2_files.endswith(".mzML"):
                mzml_w_ms2.append(ms2_files)
            else:
                for ms2_file in search_for_mzml(ms2_files):
                    mzml_w_ms2.append(ms2_file)

        if scan_experiment:
            for acq in self.experiment.acquisitions:
                if acq.has_ms2:
                    mzml_w_ms2.append(acq.mzml_filepath)
        
        matched_ms2_objects = 0
        for ms2_object in lazy_extract_ms2_spectra(mzml_w_ms2):
            used_khipu = set()
            matching_features = self.search_for_feature(
                ms2_object.prec_mz,
                ms2_object.rtime,
                mapping_mz_tol * 2,
                mapping_rt_tolerance,
            )
            for matching_feature in matching_features:
                kp_id = self.feature_id_to_khipu_id[matching_feature]
                khipu = self.dict_empcpds[kp_id]
                if "MS2_Spectra" not in khipu:
                    khipu["MS2_Spectra"] = []
                if kp_id not in used_khipu:
                    matched_ms2_objects += 1
                    khipu["MS2_Spectra"].append(ms2_object.embedding())
                    used_khipu.add(kp_id)

    @staticmethod
    def construct_from_feature_table(
        experiment,
        isotopes=None,
        adducts=None,
        ext_adducts=None,
        feature_table_moniker="full",
        moniker="default",
        add_singletons=False,
        rt_search_window=2,
        mz_tol=5,
        charges=None,
    ):
        """
        For a given feature table, generate the empirical compounds
        for that table using a set of isotopes, adducts, charges,
        and save it as either the table moniker or a new moniker.

        Args:
            isotopes (list, optional): isotopes for which to search
            adducts (list, optional): adducts to use, if None use defaults based on
                ionization.
            extended_adducts (list, optional): extended_adducts to use, if None, the
                default extended_adducts are used.
            feature_table_moniker (str, optional): the feature table to use
            empCpd_moniker (str, optional): the moniker to save the empcpds to
            :param add_singletons (bool, optional): if true, add singletons to the khipus
            rt_search_window (int, optional): the rt window to use for empcpd
                construction, default is 2.
            mz_tol: the mz tolerance in ppm to use for
                empcpd construction, default is 5.
            charges: the charges, in absolute units, to consider
                for empcpd construction.

        Returns:
            empCpd: empcpd object
        """
        charges = [1,2,3] if charges is None else charges
        ext_adducts = extended_adducts if ext_adducts is None else ext_adducts
        isotopes = isotope_search_patterns if isotopes is None else isotopes
        default_adducts = {
            "pos": adduct_search_patterns,
            "neg": adduct_search_patterns_neg
        }
        if adducts is None:
            adducts = default_adducts[experiment.ionization_mode]

        peaklist = read_table_to_peaks(
            experiment.feature_tables[feature_table_moniker],
            has_header=True,
            mz_col=1,
            rtime_col=2,
            feature_id=0,
            full_extract=True,
        )
        for p in peaklist:
            p["id"] = p["id_number"]
            p["representative_intensity"] = None

        ECCON = epdsConstructor(peaklist, experiment.ionization_mode)
        dict_empcpds = ECCON.peaks_to_epdDict(
            isotopes,
            adduct_search_patterns,
            ext_adducts,
            mz_tolerance_ppm=mz_tol,
            rt_tolerance=rt_search_window,
            charges=charges,
        )
        all_feature_ids = set()
        for empcpd in dict_empcpds.values():
            for peak in empcpd["MS1_pseudo_Spectra"]:
                all_feature_ids.add(peak["id_number"])
        if add_singletons:
            for peak in peaklist:
                if peak["id_number"] not in all_feature_ids:
                    peak["ion_relation"] = None
                    dict_empcpds[len(dict_empcpds)] = {
                        "interim_id": len(dict_empcpds),
                        "neutral_formula_mass": "",
                        "neutral_formula": "",
                        "MS1_pseudo_Spectra": [peak],
                        "MS2_spectra": []
                    }
        empcpd = EmpCpds(dict_empcpds, experiment, moniker)
        empcpd.save()
        return empcpd

    def l4_annotate(self, annotation_sources, rt_tolerance=5):
        """
        Given multiple annotation sources in the JSON format compliant with JMS, annotate based on neutral formula
        match to the annotation sources.

        :param annotation_sources: list of filepaths to annotation sources in JSON format
        :param rt_tolerance: the rt_toleance to be used by ExperimentalEcpdDatabase. Defaults to 5.
        """
        ion_mode = self.experiment.ionization_mode
        EED = ExperimentalEcpdDatabase(mode=ion_mode, rt_tolerance=rt_tolerance)
        EED.build_from_list_empCpds(list(self.dict_empcpds.values()))

        formula_entry_lookup = {}
        for source in annotation_sources:
            with open(source, encoding='utf-8') as source_fh:
                source_data = json.load(source_fh)
                for entry in source_data:
                    if entry["neutral_formula"] not in formula_entry_lookup:
                        formula_entry_lookup[entry["neutral_formula"]] = []
                    formula_entry_lookup[entry["neutral_formula"]].append(entry)
                KCD = knownCompoundDatabase()
                KCD.mass_index_list_compounds(source_data)
                KCD.build_emp_cpds_index()
                EED.extend_empCpd_annotation(KCD)

        for khipu in self.dict_empcpds.values():
            if "Level_4" not in khipu:
                khipu["Level_4"] = []
            for match in khipu.get("list_matches", []):
                formula_mass, _, _ = match
                formula, _ = formula_mass.split("_")
                if formula in formula_entry_lookup:
                    khipu["Level_4"].extend(formula_entry_lookup[formula])
        self.update_annotations()

    def l2_annotate(
        self,
        msp_files,
        mz_tol=5,
        similarity_method="CosineHungarian",
        min_peaks=1,
        score_cutoff=0.50,
    ):
        """
        This method add l2 annotations to empirical compounds. This requires that first ms2 spectra
        be mapped to the empcpd object.

        Level 2 annotations are lower confidence that Level 1 annotations but generated in a similar
        manner, MS2 similarity, but the references spectra are from a public reference database.

        The similarity method can be any method that is provided by matchms. CosineHungarian is the 
        default as it is a mathematically sound formulation of the cosine similarity and fast enough
        to be practical. 

        Args:
            msp_files (str): path to directory with ms2 mzml files
            mz_tol (int, optional): mz tolerance in ppm for the precursor_mz_match. Defaults to 5.
            similarity_method (str, optional): name of the method for the similarity metric. 
            Defaults to "CosineHungarian".
            min_peaks (int, optional): the minimum number of matching peaks between experimental 
            and reference specturm. Defaults to 2.
            score_cutoff (float, optional): the minimum score required for an annotation. 
            Defaults to 0.50.
        """        
        match = 0
        msp_files = [msp_files] if isinstance(msp_files, str) else msp_files
        similarity_method = get_similarity_method(similarity_method)
        precursor_mz_tree = self.get_precursor_mz_tree(2 * mz_tol)
        for db_ms2 in lazy_extract_ms2_spectra(msp_files, mz_tree=precursor_mz_tree):
            match_tol = db_ms2.prec_mz / 1e6 * mz_tol * 2
            sim_instance = similarity_method(tolerance=match_tol)
            for possible_match in {x.data for x in precursor_mz_tree.at(db_ms2.prec_mz)}:
                for exp_ms2 in self.ms2_spectra[possible_match]:
                    sim_result = sim_instance.pair(db_ms2.matchms_spectrum, exp_ms2.matchms_spectrum)
                    score, n_matches = sim_result.tolist()
                    if score >= score_cutoff and n_matches >= min_peaks:
                        match += 1
                        exp_ms2.annotate(
                            db_ms2,
                            score,
                            n_matches,
                            annotation_level="Level_2"
                        )
        self.update_annotations(update_ms2=True)

    def l1a_annotate(
        self,
        standards_csv,
        mz_tol=5,
        rt_tolerance=30,
        similarity_method="CosineHungarian",
        min_peaks=1,
        score_cutoff=0.50,
    ):
        """
        Perform l1 annotation on the empcpds. Using CD authentic standard library. 

        Args:
            standards_csv (str): path to CD csv export
            mz_tol (int, optional): mz tolerance to match precursors. Defaults to 5. 
            Note that this value will be multiplied by 2
            rt_tolerance (int, optional): rt tolerance to match precursors. Defaults to 30.
            similarity_method (str, optional): which matchms similarity method to use. Defaults to "CosineHungarian".
            min_peaks (int, optional): minimum number of peaks that must be shared for annotation. Defaults to 2.
            score_cutoff (float, optional): scores above this value are consider matches. Defaults to 0.50.
        """
        similarity_method = get_similarity_method(similarity_method)
        precursor_mz_tree = self.get_precursor_mz_tree(2 * mz_tol)
        precursor_rt_tree = self.get_precursor_rt_tree(rt_tolerance)
        for db_ms2 in extract_CD_csv(standards_csv, self.experiment.ionization_mode, lazy=True):
            match_tol = db_ms2.prec_mz / 1e6 * mz_tol * 2
            sim_instance = similarity_method(tolerance=match_tol)
            for possible_match in {x.data for x in precursor_mz_tree.at(db_ms2.prec_mz)}:
                if possible_match in {x.data for x in precursor_rt_tree.at(db_ms2.rtime)}:
                    for exp_ms2 in self.ms2_spectra[possible_match]:
                        sim_res = sim_instance.pair(db_ms2.matchms_spectrum, exp_ms2.matchms_spectrum)
                        score, n_matches = sim_res.tolist()
                        if score >= score_cutoff and n_matches >= min_peaks:
                            exp_ms2.annotate(
                                db_ms2,
                                score,
                                n_matches,
                                annotation_level="Level_1a",
                            )
        self.update_annotations(update_ms2=True)

    def l1b_annotate(self, standards_csv, mz_tol=5, rt_tolerance=10):
        """
        Level1b annotations are based on mz, rtime tolerance against known standards. 

        This method takes the exported standard library from mz vault and compares a feature's 
        rtime and mz to the standard's mz and retention time. 

        Args:
            standards_csv (str): path to mzvault export
            mz_tol (int, optional): mz tolerance in ppm. Defaults to 5.
            rt_tolerance (int, optional): rt tolerance in seconds. Defaults to 10.
        """
        for csv in standards_csv:
            for standard in pd.read_csv(csv).to_dict(orient='records'):
                mz, rtime, cname = [
                    standard[k] for k in ["Confirm Precursor", "RT", "CompoundName"]
                ]
                for feature_match in self.search_for_feature(
                    mz, rtime, mz_tol, rt_tolerance
                ):
                    kp = self.dict_empcpds[self.feature_id_to_khipu_id[feature_match]]
                    if "Level_1b" not in kp:
                        kp["Level_1b"] = []
                    kp["Level_1b"].append((cname, csv))
        self.update_annotations()




import pandas as pd
import yaml
import warnings
import json
from metDataModel.core import EmpiricalCompound

def read_metadata_file(input_file_path):
    '''
    Read a metadata YAML file.

    Input
    =====
    the input file path

    An example of a qualified YAML file
    ======================
    ion_relationship: '[M+H]+'
    mode: 'HILICpos'
    description: 'HILICpos authentic standard library from Shujian and Minghao'
    curator: 'Shujian Zheng|Minghao Gong'
    method: '5 min method'
    notes:
    initiation:
        curator: MG
        date: ''
        note: '5 min method, the library was created and combined by Minghao Gong and Shujian Zheng'

    version: '0.1'
    label: 'combn_1strnd'

    Output
    ======
    A dictionary of the metadata
    '''
    with open(input_file_path, 'r') as f:
        metadata = yaml.safe_load(f)
    return(metadata)
    
def libTab2EpdJson(df_path = '', delimiter = ', ',metadata = '', RT_long_format = True):
    '''
    Input
    =====
    df_path: the input tabular file path
    delimiter: the delimiter used in separate values in individual cells (note: this delimiter is not for separation between cells) 
    metadata: see also read_metadata_file
    RT_long_format: If there are incidents multiple retention time for a single compound stored in individual row

    Note
    ====
    the provided table needs to be comma separated data
    the metadata needs to be in yaml format
    The columns need to follow the rule as: 'mz','rt','identifier','name','Formula','adduct'
    If multiple adducts/mz of a particular empirical compound present, it needs to be at the same row (retention time is always the same with different adducts)
    The number of delimited items in identifier, name & Formula should be the same.

    Example of input tabular data
    =============================
    mz	rt	identifier	name	Formula	adduct
	130.0498696	231.24	HMDB0000805, HMDB0000267	D-Pyroglutamic acid, L-Pyroglutamic acid	C5H7NO3, C5H7NO3	[M+H]+
    130.0862551	127.38	HMDB0000716	L(-)-Pipecolinic acid	C6H11NO2	[M+H]+
    130.0862551	230.16	HMDB0000716	L(-)-Pipecolinic acid	C6H11NO2	[M+H]+
    132.0655197	154.14	HMDB0060460, HMDB0000725	cis-4-Hydroxy-D-proline, L-Hydroxyproline	C5H9NO3, C5H9NO3	[M+H]+
    132.0655197	219.3	HMDB0060460, HMDB0000725	cis-4-Hydroxy-D-proline, L-Hydroxyproline	C5H9NO3, C5H9NO3	[M+H]+

    Output
    ======
    a dictionary contains a list of empirical compound and a metadata

    '''
    df = pd.read_csv(df_path, dtype=str)

    if RT_long_format == True:
        reshaped_df = df.groupby([x for x in df.columns if 'rt'not in x]).agg({'rt': lambda x: ', '.join(str(i) for i in x)})
        reshaped_df = reshaped_df.reset_index()
    else:
        reshaped_df = df
    myEpds = []
    final_dict = reshaped_df.to_dict('index')
    
    for row in final_dict.values():
        Epd = EmpiricalCompound()
        
        # dealing with identity
        if delimiter in row['identifier']:
            ids = row['identifier'].split(delimiter)
            names = row['name'].split(delimiter)
            formulas = row['Formula'].split(delimiter)
        else:
            ids = [row['identifier']]
            names = [row['name']]
            formulas = [row['Formula']]
        for _id, name, formula in zip(ids,names,formulas):
            Epd.identity.append({'compounds': _id, 'names': name, 'formulas':formula, 'score': 0, 'probability': 0})
                   
        # dealing with MS1_pseudo_Spectra
        if delimiter in row['rt']:
            rts = row['rt'].split(delimiter)
        else:
            rts = [row['rt']]
        if delimiter in row['mz']:
            mzs = row['mz'].split(delimiter)
            adducts = row['adduct'].split(delimiter)
        else:
            mzs = [row['mz']]
            adducts = [row['adduct']]
        for rt in rts:
            for mz,adduct in zip(mzs,adducts):
                Epd.MS1_pseudo_Spectra.append({'feature_id': '', 
                                               'mz': float(mz.rstrip()), 
                                           'rtime': float(rt.rstrip()), 'charged_formula': '', 'ion_relation': adduct},)
        myEpds.append(Epd)
    
    Epd_dict = {}
    Epd_dict['list_of_Empirical_Compounds'] = [x.__dict__ for x in myEpds]
    Epd_dict['metadata'] = metadata

    return(Epd_dict)

def write_json(EpdJson_dict,output_file_path):
    '''
    write the empirical compound dictionary into output file under the correct path.
    '''
    with open(output_file_path, "w") as f:
        json.dump(EpdJson_dict,f,indent = 2)

def combn_lib(lib1,lib2,metadata = ''):
    '''
    Combine two library Json empirical compound dictionary 

    Input
    =====
    Two Empirical compound dictionary (the dictionary includes a list of empirical compound and a previous metadata)
    A metadata YAML file (See also read_metadata_file)

    Output
    ======
    Output a merged Empirical compound json file, attached with new metadata based on YAML file
    If there are overlaps of compound identifiers between two json files, create a warning message for further checking

    '''
    lib1_cpd_id = [cpd['compounds'] for epd in lib1['list_of_Empirical_Compounds'] for cpd in epd['identity']]
    lib2_cpd_id = [cpd['compounds'] for epd in lib2['list_of_Empirical_Compounds'] for cpd in epd['identity']]
    intersect = set(lib1_cpd_id).intersection(set(lib2_cpd_id))
    
    lib3 = {'list_of_Empirical_Compounds':  [], 'metadata' : metadata}
    if len(intersect) !=0:
        intersect_string = ";".join(list(intersect))
        warnings.warn(f"There are duplicated identifiers. Check for duplications of {intersect_string}")
    lib3['list_of_Empirical_Compounds'] = lib1['list_of_Empirical_Compounds'] + lib2['list_of_Empirical_Compounds']

    return(lib3)

   


'''
Experiment.py contains the pipeline's implementation of the Experiment object which in turn 
inherits from the Experiment object in the metdatamodel. 

The Experiment is largely a computational aid and orchestrates an analysis. This includes
keeping track of intermediate results and acquisitions. 

'''

import os
import random
import sys
import time
import json
import subprocess
import multiprocessing as mp
import pandas as pd
import matplotlib.colors as mcolors
import asari
from metDataModel import core
from .utils import recursive_encoder, file_operations, flatten_nested_dicts
from . import Acquisition
from . import EmpCpds
from . import FeatureTable

class Experiment(core.Experiment):
    """
    The experiment object represents a set of acquisitions.

    This super vague constructor was useful during testing, now
    will explicitly define all the fields.
    """


    # this needs to be removed but determines where intermediates are stored
    subdirectories = {
        "converted_subdirectory": "converted_acquisitions/",
        "raw_subdirectory": "raw_acquisitions/",
        "acquisition_data": "acquisitions/",
        "annotation_subdirectory": "annotations/",
        "filtered_feature_tables_subdirectory": "filtered_feature_tables/",
        "ms2_directory": "ms2_acquisitions/",
        "qaqc_figs": "QAQC_figs/",
        "asari_subdirectory": "asari/",
        "output_subdirectory": "output/"
    }

    # need to minimize parameters
    def __init__(
        self,
        experiment_name,
        experiment_directory,
        acquisitions=None,
        qcqa_results=None,
        feature_tables=None,
        empCpds=None,
        log_transformed_feature_tables=None,
        ionization_mode=None,
        cosmetics=None,
        used_cosmetics=None,
        MS2_methods=None,
        MS1_only_methods=None,
        command_history=None,
        study=None,
        sequence=None,
        final_empCpds=None,
        species=None,
        tissue=None,
        provenance=None,
        converted_subdirectory=None,
        raw_subdirectory=None,
        acquisition_data=None,
        annotation_subdirectory=None,
        filtered_feature_tables_subdirectory=None,
        ms2_directory=None,
        qaqc_figs=None,
        asari_subdirectory=None,
        output_subdirectory=None,
        ordered_samples=None,
        parent_study=None
    ):
        super().__init__(id=experiment_name,
                         parent_study= parent_study if parent_study else '',
                         species=species if species else '',
                         tissue=tissue if tissue else '',
                         provenance=provenance if provenance else {
                                'generated_time': '',
                                'generated_by': '',
                                'input_filename': '',
                                'preprocess_software': '',
                                'preprocess_parameters': {}
                            },
                         chromatography= {
                                'type': '',
                                'total_time': '',
                                'method_file': '',
                                'column_model': '',
                                'column_length': ''
                            },
                         instrumentation= {
                                'type': '',
                                'spectrometer': '',
                                'method_file': '',
                                'ionization': ''
                            },
                         ObservationAnnotation= {
                                'sample_list': [],
                                'file_sample_mapper': {}
                            },
                        #ordered_samples = ordered_samples if ordered_samples else [x.acq_name for x in acquisitions] if acquisitions else []
                         )

        self.experiment_directory = os.path.abspath(experiment_directory)
        self.acquisitions = acquisitions if acquisitions else []
        self.qcqa_results = qcqa_results if qcqa_results else {}
        self.feature_tables = feature_tables if feature_tables else {}
        self.empCpds = empCpds if empCpds else {}
        self.log_transformed_feature_tables = (
            list(set(log_transformed_feature_tables)) if log_transformed_feature_tables else []
        )
        self.cosmetics = cosmetics if cosmetics else {}
        self.used_cosmetics = used_cosmetics if used_cosmetics else {}
        self.MS2_methods = MS2_methods if MS2_methods else set()
        self.MS1_only_methods = MS1_only_methods if MS1_only_methods else set()
        self.command_history = command_history if command_history else []
        self.sequence = sequence
        self.List_of_empCpds = self.final_empCpds = final_empCpds

        # subdirectories
        # todo - someday, lets make this more clean
        self.converted_subdirectory = converted_subdirectory
        self.raw_subdirectory = raw_subdirectory
        self.acquisition_data = acquisition_data
        self.annotation_subdirectory = annotation_subdirectory
        self.filtered_feature_tables_subdirectory = filtered_feature_tables_subdirectory
        self.ms2_directory = ms2_directory
        self.qaqc_figs = qaqc_figs
        self.asari_subdirectory = asari_subdirectory
        self.output_subdirectory = output_subdirectory

        # special metadata fields
        self.species = species
        self.tissue = tissue

        # lazily evaluated parameters
        self.__ionization_mode = (
            ionization_mode  # this one can be costly, lets store it
        )
        self.__acq_names = None
        self.__ms2_acquisitions = None

    @property
    def experiment_name(self):
        return self.id
    
    @property
    def study(self):
        return self.parent_study

    def order_samples(self):
        """
        This updates the ordered_samples param, part of metdatamodel implementation.
        """
        self.ordered_samples = [x.name for x in self.acquisitions]

    @staticmethod
    def create_experiment(experiment_name, experiment_directory, sequence=None):
        """
        This is the main constructor for an experiment object.

        This requires a name for the experiment, the directory to which
        to write intermediates and optionally the ionization mode.

        Args:
            experiment_name (str): a moniker for the experiment
            experiment_directory (str): if true, return the object else its path. Defaults to False.
            ionization_mode (str): the ionization mode of the acquisitions can be 'pos', 'neg', or None for auto-detection.

        Returns:
            experiment object
        """
        to_create = [os.path.abspath(experiment_directory)]
        for subdir in Experiment.subdirectories.values():
            path = os.path.join(to_create[0], subdir)
            to_create.append(path)
        for subdirectory_full_path in to_create:
            os.makedirs(subdirectory_full_path, exist_ok=True)
        full_subdirs = dict(zip(Experiment.subdirectories.keys(), to_create[1:]))
        return Experiment(
            experiment_name,
            os.path.abspath(experiment_directory),
            ionization_mode=None,
            command_history=[str(time.time()) + ":start_analysis"],
            sequence=sequence,
            converted_subdirectory=full_subdirs["converted_subdirectory"],
            raw_subdirectory=full_subdirs["raw_subdirectory"],
            acquisition_data=full_subdirs["acquisition_data"],
            annotation_subdirectory=full_subdirs["annotation_subdirectory"],
            filtered_feature_tables_subdirectory=full_subdirs[
                "filtered_feature_tables_subdirectory"
            ],
            ms2_directory=full_subdirs["ms2_directory"],
            qaqc_figs=full_subdirs["qaqc_figs"],
            asari_subdirectory=full_subdirs["asari_subdirectory"],
            output_subdirectory=full_subdirs["output_subdirectory"]
        )

    def save(self):
        """
        This saves the experiment object as a JSON object in side the
        experiment directory.
        """
        try:
            save_path = os.path.join(self.experiment_directory, "experiment.json.tmp")
            with open(os.path.join(save_path), "w", encoding="utf-8") as save_filehandle:
                self.command_history.append(str(time.time()) + ":" + ";".join(sys.argv))
                json.dump(recursive_encoder(self.__dict__), save_filehandle, indent=4)
                file_operations["move"](save_path, save_path.replace(".tmp", ""))
        except Exception as e:
            print("Unable to save experiment!\n" + e)
            raise e

    @staticmethod
    def load(experiment_json_filepath):
        """
        Reconstitute the experiment object from a saved JSON file representing the object

        Args:
            experiment_json_filepath (str): path to the JSON file
        
        Returns:
            an experiment object

        """
        with open(experiment_json_filepath, encoding="utf-8") as json_filehandle:
            decoded_JSON = json.load(json_filehandle)
            experiment = Experiment(
                decoded_JSON["id"],
                decoded_JSON["experiment_directory"],
                None,
                qcqa_results=decoded_JSON["qcqa_results"],
                feature_tables=decoded_JSON["feature_tables"],
                empCpds=decoded_JSON["empCpds"],
                log_transformed_feature_tables=decoded_JSON[
                    "log_transformed_feature_tables"
                ],
                ionization_mode=decoded_JSON["_Experiment__ionization_mode"],
                cosmetics=decoded_JSON["cosmetics"],
                used_cosmetics=decoded_JSON["used_cosmetics"],
                MS2_methods=decoded_JSON["MS2_methods"],
                MS1_only_methods=decoded_JSON["MS1_only_methods"],
                command_history=decoded_JSON["command_history"],
                study=None,
                sequence=decoded_JSON["sequence"],
                final_empCpds=decoded_JSON["final_empCpds"],
                species=decoded_JSON["species"],
                tissue=decoded_JSON["tissue"],
                provenance=decoded_JSON["provenance"],
                converted_subdirectory=decoded_JSON["converted_subdirectory"],
                raw_subdirectory=decoded_JSON["raw_subdirectory"],
                acquisition_data=decoded_JSON["acquisition_data"],
                annotation_subdirectory=decoded_JSON["annotation_subdirectory"],
                filtered_feature_tables_subdirectory=decoded_JSON[
                    "filtered_feature_tables_subdirectory"
                ],
                ms2_directory=decoded_JSON["ms2_directory"],
                qaqc_figs=decoded_JSON["qaqc_figs"],
                asari_subdirectory=decoded_JSON["asari_subdirectory"],
                output_subdirectory=decoded_JSON['output_subdirectory'],
                parent_study=decoded_JSON['parent_study']
            )
        for x in decoded_JSON["acquisitions"]:
            experiment.acquisitions.append(
                Acquisition.Acquisition.load_acquisition(x, experiment)
            )
        experiment.order_samples()
        return experiment

    @property
    def ms2_acquisitions(self):
        """
        This returns all acquisitions in the experiment that have MS2.
        Lazily evaluated.

        Returns:
            list of acquisitions with MS2
        """
        if self.__ms2_acquisitions is None:
            self.__ms2_acquisitions = [acq for acq in self.acquisitions if acq.has_ms2]
        return self.__ms2_acquisitions

    @property
    def sample_names(self):
        """
        This returns the name of all acquisitions in the experiment
        Lazily evaluated.

        Returns:
            names of all samples in the experiment
        """
        if self.__acq_names is None:
            self.__acq_names = [acq.name for acq in self.acquisitions]
        return self.__acq_names

    @property
    def ionization_mode(self):
        """
        This returns the user-specified or determined ionization mode
        of the experiment's acquisitions.
        Lazily evaluated.

        Returns:
            the ionization mode 'pos' or 'neg'
        """
        ion_modes = []
        if self.__ionization_mode is None:
            while len(ion_modes) < min(3, len(self.acquisitions)):
                acquisition = random.sample(self.acquisitions, 1)[0]
                if acquisition.ionization_mode:
                    ion_modes.append(acquisition.ionization_mode)
            if len(set(ion_modes)) == 1:
                self.__ionization_mode = list(ion_modes)[0]
        return self.__ionization_mode

    def delete_feature_table(self, moniker):
        """
        This method will safely delete a feature table and unregister it with the experiment.

        Args:
            moniker (str): the table moniker to delete

        """
        if moniker == "*":
            for moniker in list(self.feature_tables.keys()):
                if moniker not in {"preferred", "full"}:
                    self.delete_feature_table(moniker)
        elif moniker in list(self.feature_tables.keys()):
            file_operations["delete"](self.feature_tables[moniker])
            del self.feature_tables[moniker]
        else:
            print("No such table: ", moniker)

    def delete_empCpds(self, moniker):
        """
        This method will safely delete an empcpd and unregister it with the experiment.

        Args:
            moniker (str): the empcpd moniker to delete

        """
        if moniker == "*":
            for moniker in list(self.empCpds.keys()):
                if moniker not in {"asari"}:
                    self.delete_empCpds(moniker)
        elif moniker in list(self.empCpds.keys()):
            file_operations["delete"](self.empCpds[moniker])
            del self.empCpds[moniker]
        else:
            print("No such empCpds: ", moniker)

    def retrieve_feature_table(self, moniker, as_object=False):
        """
        For a given moniker return either the feature table object or its path.

        Args:
            moniker (str): the table to retrieve
            as_object (bool, optional): if true, return the object else its path. Defaults to False.

        Returns:
            str or object: the feature table or its path
        """
        if moniker in self.feature_tables:
            if as_object:
                return FeatureTable.FeatureTable.load(moniker, self)
            return self.feature_tables[moniker]
        print("No such table: ", moniker)
        sys.exit()

    def retrieve_empCpds(self, moniker, as_object=False):
        """
        For a given moniker return either the empcpd object or its path.

        Args:
            moniker (str): the empcpd to retrieve
            as_object (bool, optional): if true, return the object else its path. Defaults to False.

        Returns:
            str or object: the feature table or its path
        """
        if moniker in self.empCpds:
            if as_object:
                return EmpCpds.EmpCpds.load(moniker, self)
            return self.empCpds[moniker]
        print("No such empCpds: ", moniker)
        sys.exit()

    def create_sample_annotation_table(self):
        """
        Create the sample annotation table which maps samples to their metadata. 

        Returns:
            dataframe: the table as a dataframe
        """
        annotation_table = []
        for acquisition in self.acquisitions:
            flat_acq_dict = flatten_nested_dicts(acquisition.__dict__)
            to_delete = []
            for key in flat_acq_dict:
                if key.startswith('_Acquisition__'):
                    to_delete.append(key)
            for key_to_delete in to_delete:
                del flat_acq_dict[key_to_delete]
            annotation_table.append(flat_acq_dict)
        return pd.DataFrame(annotation_table)

    def add_acquisition(self, acquisition, mode="link"):
        """
        This method adds an acquisition to the list of acquisitions in the experiment, ensures 
        there are no duplicates and then links or copies the acquisition, currently only as a 
        .raw file, to the experiment directory

        Args:

        acquisition (object): an Acquistiion object
        mode (str): how to move acquisitions into the experiment, default "link", can be "copy"
        method_field (str): this is the field to check for the method name, used to shortcircuit 
            MS2 determination
        """
        if os.path.exists(acquisition.source_filepath):
            target_path = None
            source_basepath = os.path.basename(acquisition.source_filepath)
            if acquisition.source_filepath.endswith(".mzML"):
                acquisition.raw_filepath = None
                if acquisition.has_ms2:
                    target_path = os.path.join(self.ms2_directory, source_basepath)
                else:
                    target_path = os.path.join(
                        self.converted_subdirectory, source_basepath
                    )
                acquisition.mzml_filepath = target_path
            elif acquisition.source_filepath.endswith(".raw"):
                target_path = os.path.join(
                    self.raw_subdirectory, source_basepath
                )
                acquisition.raw_filepath = target_path 
            if target_path is not None and not os.path.exists(target_path):
                file_operations[mode](acquisition.source_filepath, target_path)
                self.acquisitions.append(acquisition)
            else:
                print(
                    "Target already exists: ",
                    acquisition.name,
                    acquisition.source_filepath,
                )
        else:
            print("File Not Found: ", acquisition.name, acquisition.source_filepath)
        self.order_samples()
        if self.number_samples is None:
            self.number_samples = 0
        self.number_samples += 1
        if self.species is None:
            self.species = set()
        elif self.species is not set():
            self.species = set(self.species)
        self.species.add(acquisition.metadata_tags.get('species', 'Unknown'))
        self.species = list(self.species)
        if self.tissue is None:
            self.tissue = set()
        elif self.tissue is not set():
            self.tissue = set(self.tissue)
        self.tissue.add(acquisition.metadata_tags.get('species', 'Unknown'))
        self.tissue = list(self.tissue)

    def generate_output(self, empCpd_moniker, table_moniker, comprehensive_output=False):
        """
        This generates and stores the the feature table, sample annotation table, and the 
        feature annotation table to the output directory. It also copies the JSON for the 
        desried empcpd and experiment to the directory. 

        Args:
            empCpd_moniker (str): moniker of empcpd to use
            table_moniker (str): moniker of table to use
        """
        feature_table_path = os.path.join(self.output_subdirectory, "feature_table.tsv")
        feature_table = self.retrieve_feature_table(table_moniker, True)
        feature_table.feature_table.to_csv(feature_table_path, sep="\t", index=False)

        sample_annotation_table_path = os.path.join(self.output_subdirectory, "sample_annot_table.tsv")
        sample_annotation_table = self.create_sample_annotation_table()
        sample_annotation_table.to_csv(sample_annotation_table_path, sep="\t", index=False)

        annotation_table_path = os.path.join(self.output_subdirectory, "annotation_table.tsv")
        empCpds = self.retrieve_empCpds(empCpd_moniker, True)
        annotation_table = empCpds.create_annotation_table(comprehensive_output)
        annotation_table.to_csv(annotation_table_path, sep="\t", index=False)

        file_operations["copy"](self.retrieve_empCpds(empCpd_moniker, False), self.output_subdirectory)
        file_operations["copy"](self.retrieve_feature_table(table_moniker, False), self.output_subdirectory)
        file_operations["copy"](os.path.join(self.experiment_directory, "experiment.json"), self.output_subdirectory)

    def convert_raw_to_mzML(self, conversion_command, num_cores=4):
        """
        Convert all raw files to mzML

        Args

        conversion_command (str or list): This specifies the command to call
            to perform the conversion. Can be list or space-delimited
            string. Must contain $RAW_PATH and $OUT_PATH where the
            input and output file names will go.
        """
        jobs = []
        output_paths = []
        raw_acquisitions = [
            acq for acq in self.acquisitions if acq.mzml_filepath is None
        ]
        for acquisition in raw_acquisitions:
            output_name = os.path.basename(acquisition.raw_filepath).replace(
                ".raw", ".mzML"
            )
            output_filepath = os.path.join(self.converted_subdirectory, output_name)
            output_paths.append(output_filepath)
            field_map = {
                "$RAW_PATH": acquisition.raw_filepath,
                "$OUT_PATH": output_filepath,
            }
            if isinstance(conversion_command, list):
                jobs.append(
                    [field_map.get(element, element) for element in conversion_command]
                )
            elif isinstance(conversion_command, str):
                for field, replacement_value in field_map.items():
                    conversion_command = conversion_command.replace(
                        field, replacement_value
                    )
        with mp.Pool(num_cores) as workers:
            for _, acquisition, output_path in zip(
                workers.map(os.system, [" ".join(x) for x in jobs]),
                raw_acquisitions,
                output_paths,
            ):
                acquisition.mzml_filepath = output_path
                if acquisition.has_ms2:
                    ms2_path = os.path.join(
                        self.ms2_directory, os.path.basename(acquisition.mzml_filepath)
                    )
                    file_operations["move"](acquisition.mzml_filepath, ms2_path)
                    acquisition.mzml_filepath = ms2_path

    def filter_samples(self, sample_filter, return_field=None):
        """
        Find the set of acquisitions that pass the provided filter and return either the 
        acquisition object or the specified field of each passing sample

        Args:

        filter (dict): a filter dictionary
        return_field (str, optional): if provided and valid
            the field specified is return per matching acquisition.

        :return: list of matching acquisitions or the return_field value of the acquisitions
        """
        if return_field:
            return [
                getattr(acquisition, return_field)
                for acquisition in self.acquisitions
                if acquisition.filter(sample_filter)
            ]
        return [
            acquisition
            for acquisition in self.acquisitions
            if acquisition.filter(sample_filter)
        ]

    @staticmethod
    def construct_experiment_from_CSV(
        experiment_directory,
        csv_filepath,
        sample_filter=None,
        name_field="File Name",
        path_field="Filepath",
        sample_skip_list_fp=None,
        file_mode="link"
    ):
        """
        For a given sequence file, create the experiment object, and add all acquisitions

        Args:
            experiment_directory (str): path to store experiment and intermediates
            csv_filepath (str): filepath to sequence CSV
            ionization (str, optional): default None, can be 'pos' or 'neg'. The ionization mode of
                the experiment. If None, it will be determined automatically
            filter (dict, optional): a filter dictionary, only matching entries are included
            name_field (str, optional): the column from which to extract the acquisition name
            path_field (str, optional): the column from which to extract the acquisition filepath
            sample_skip_list (str, optional): path to a txt file with sample names to exclude

        Returns:
            experiment object
        """
        acq_constructor = Acquisition.Acquisition.create_acquisition
        sample_skip_list = set()
        if sample_skip_list_fp:
            if sample_skip_list_fp.endswith(".txt"):
                with open(sample_skip_list_fp, encoding='utf-8') as skip_list_fh:
                    sample_skip_list = {x.strip() for x in skip_list_fh.readlines()}

        if not os.path.exists(os.path.join(experiment_directory, "experiment.json")):
            if isinstance(sample_filter, dict):
                pass
            elif isinstance(sample_filter, str) and sample_filter:
                sample_filter = json.loads(sample_filter)
                assert isinstance(sample_filter, dict)
            elif sample_filter is None:
                sample_filter = {}
            experiment = Experiment.create_experiment(
                experiment_directory, experiment_directory, sequence=csv_filepath
            )
            sequence_df = pd.read_csv(csv_filepath)
            sequence_df = sequence_df.infer_objects(copy=True)
            for acq_info in sequence_df.to_dict(orient="records"):
                acq_path = acq_info.get(path_field, None)
                acq_name = acq_info.get(name_field, None)
                if acq_path is None and acq_name:
                    print("Acquisition path not found, looking in local directory...")
                    csv_directory = os.path.dirname(csv_filepath)
                    possible_mzml_path = os.path.join(csv_directory, acq_name + ".mzML")
                    possible_raw_path = os.path.join(csv_directory, acq_name + ".raw")
                    if os.path.exists(possible_mzml_path):
                        print("\tmzml found")
                        acq_path = possible_mzml_path
                    elif os.path.exists(possible_raw_path):
                        print("\traw found")
                        acq_path = possible_raw_path
                if acq_name and acq_path:
                    acquisition = acq_constructor(acq_name, acq_path, acq_info)
                    acquisition.experiment = experiment
                    if acquisition.filter(sample_filter):
                        if acquisition.name not in sample_skip_list:
                            experiment.add_acquisition(acquisition, mode=file_mode)
                else:
                    print("Skipping: ", acq_name, " no acquisition data found")
            if experiment.acquisitions:
                experiment.save()
                return experiment
            file_operations["delete"](experiment.experiment_directory)
            print("Unable to create empty experiment")
            sys.exit()
        else:
            print("Experiment already exists!")
            sys.exit()

    def summarize(self):
        """
        Print the list of empCpds and feature tables in the experiment to the console
        """
        print("empCpds:")
        for moniker, path in self.empCpds.items():
            print("\t", moniker, " - ", path)
        print("feature tables:")
        for moniker, path in self.feature_tables.items():
            print("\t", moniker, " - ", path)

    def generate_cosmetic_map(self, field=None, provided_cos_type="color", seed=None):
        """
        This generates the mapping of acquisition properties to colors,
        markers, text, etc. used in figure generation. This allows for
        consistency across runs as the mapping is stored in the object.

        Args:


        field (str, optional): field for which to generate the mapping
        cos_type (str, optional): 'color', 'marker', or 'text'
        seed (int, optional): used for shuffling, by setting this value, the
            same exact mapping can be generated each time.

        Returns:
            dict: the mapping of fields to cosmetic types.
        """

        # todo - this should be in JSON
        cosmetic_rules = {
            "colors": {
                "banned": [
                    "snow",
                    "beige",
                    "honeydew",
                    "azure",
                    "aliceblue",
                    "lightcyan",
                    "lightyellow",
                    "white",
                    "oldlace",
                    "antiquewhite",
                    "ivory",
                    "whitesmoke",
                    "mistyrose",
                    "seashell",
                    "linen",
                    "antiquewhite",
                    "lightgoldenrodyellow",
                    "cornsilk",
                    "lemonchiffon",
                    "honeydew",
                    "mintcream",
                    "ghostwhite",
                    "lavenderblush",
                ],
                "allowed": mcolors.CSS4_COLORS,
            },
            "markers": {
                "banned": [],
                "allowed": [
                    ".",
                    "o",
                    "v",
                    "^",
                    ">",
                    "<",
                    "1",
                    "2",
                    "3",
                    "4",
                    "8",
                    "s",
                    "P",
                ],
            },
            "text": {"banned": [], "allowed": []},
        }
        if seed:
            random.seed(seed)
        if provided_cos_type in self.cosmetics and field in self.cosmetics[provided_cos_type]:
            return self.cosmetics[provided_cos_type][field]

        allowed_cosmetics = {}
        for cos_type, mask in cosmetic_rules.items():
            allowed_cosmetics[cos_type] = []
            for item in mask["allowed"]:
                if item not in self.used_cosmetics:
                    if item not in mask["banned"]:
                        allowed_cosmetics[cos_type].append(item)

        if provided_cos_type in allowed_cosmetics:
            random.shuffle(allowed_cosmetics[provided_cos_type])
            cosmetic_map = {}
            for i, needs_cosmetic in enumerate(
                {a.metadata_tags[field] for a in self.acquisitions}
            ):
                cosmetic_map[needs_cosmetic] = allowed_cosmetics[provided_cos_type][i]
            if provided_cos_type not in self.used_cosmetics:
                self.used_cosmetics[provided_cos_type] = []
            self.used_cosmetics[provided_cos_type].extend(list(cosmetic_map.values()))
            if provided_cos_type not in self.cosmetics:
                self.cosmetics[provided_cos_type] = {}
            self.cosmetics[provided_cos_type][field] = cosmetic_map
            self.save()
            return cosmetic_map
        print("No mapping for cosmetic type: ", provided_cos_type)
        sys.exit()

    def batches(self, batch_field):
        """
        This will group samples into 'batches', based on the user
        provided 'batch_field'.

        :param batch_field: field by which to batch samples

        :return: dictionary of batches to lists of samples
        """
        batches = {}
        for acquisition in self.acquisitions:
            batch_field_value = acquisition.metadata_tags[batch_field]
            if batch_field_value not in batches:
                batches[batch_field_value] = []
            batches[batch_field_value].append(acquisition.name)
        return batches

    def asari(self, asari_cmd, force=False):
        """
        This command will run asari on the mzml acquisitions in an
        experiment. The details of the command to be ran is defined by
        asari_cmd.

        Args:

        asari_cmd (str or list): can be string or space delimited list, must
            contain the fields $IONIZATION_MODE, $CONVERTED_SUBDIR, and
            $ASARI_SUBDIR which will be populated by this function.

            $CONVERTED_SUBDIR is where input mzml is located
            $ASARI_SUBDIR is where the results will be output
            $IONIZATION_MODE is the ionization mode of the experiment

        force (bool): if true, rerun asari if previously ran
        """

        if (
            not ("full" in self.feature_tables and "preferred" in self.feature_tables)
            or force
        ):
            mapping = {
                "$IONIZATION_MODE": self.ionization_mode,
                "$CONVERTED_SUBDIR": self.converted_subdirectory,
                "$ASARI_SUBDIR": self.asari_subdirectory,
            }
            job = asari_cmd if isinstance(asari_cmd, list) else asari_cmd.split(" ")
            job = [mapping.get(f, f) for f in asari_cmd]
            completed_process = subprocess.run(job, check=False)
            if completed_process.returncode == 0:
                for x in os.listdir(self.experiment_directory):
                    if x.startswith("asari") and "project" in x:
                        self.feature_tables.update(
                            {
                                "full": os.path.join(
                                    self.experiment_directory,
                                    x,
                                    "export/full_Feature_table.tsv",
                                ),
                                "preferred": os.path.join(
                                    self.experiment_directory,
                                    x,
                                    "preferred_Feature_table.tsv",
                                ),
                            }
                        )
                        self.empCpds.update(
                            {
                                "asari": os.path.join(
                                    self.experiment_directory,
                                    x,
                                    "Annotated_empiricalCompounds.json",
                                )
                            }
                        )
                self.provenance["preprocess_software"] = "Asari"
                if "preprocess_parameters" not in self.provenance:
                    self.provenance["preprocess_parameters"] = {}
                self.provenance["preprocess_parameters"]["version"] = asari.__version__
                for i, x in enumerate(job):
                    if x.startswith("-") or x.startswith("--"):
                        self.provenance["preprocess_parameters"][x] = job[i + 1]
            else:
                print("Error running Asari!")
                sys.exit()
        else:
            print("Asari has already been ran, pass '--force' to override")
            sys.exit()



from setuptools import setup, find_packages

with open("pcpfm/__init__.py") as f:
    exec([x for x in f.readlines() if '__version__' in x][0])

with open("README.md", "r") as f:
    long_description = f.read()

with open("requirements.txt", "r") as f:
    requirements = f.read()

setup(
  name='pcpfm',
  version=__version__,
  author='Joshua Mitchell',
  author_email='joshua.mitchell@jax.org',
  description='LC-MS metabolomics data processing pipeline',
  long_description=long_description,
  long_description_content_type="text/markdown",
  url='https://github.com/jmmitc06/PythonCentricPipelineForMetabolomics',
  license='MIT',
  keywords='metabolomics bioinformatics mass spectrometry',

  classifiers=[
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: BSD License',
    'Natural Language :: English',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3',
    'Topic :: Scientific/Engineering :: Bio-Informatics',
    'Topic :: Scientific/Engineering :: Chemistry',
    'Topic :: Software Development :: Libraries :: Python Modules',
  ],

  packages=find_packages(
    include=['*', '']
  ),
  include_package_data=True,
  #zip_safe=True,
  entry_points = {
        'console_scripts': ['pcpfm=pcpfm.main:CLI'],
    },

  python_requires='>=3.7',
  install_requires=requirements,
)